from nmigen import *

from nap.vendor import JTAG as GenericJTAG


class JTAG(GenericJTAG.implementation):
    def elaborate(self, platform):
        m = Module()
        tck = Signal(attrs={"KEEP": "TRUE"})
        platform.add_clock_constraint(tck, 1e6)
        m.domains += ClockDomain(self.jtag_domain)
        m.d.comb += ClockSignal(self.jtag_domain).eq(tck)

        shift = Signal()
        m.d.comb += self.shift_tdi.eq(shift)
        m.d.comb += self.shift_tdo.eq(shift)

        m.submodules.jtag_primitive = Instance(
            "BSCANE2",
            p_JTAG_CHAIN=1,

            o_TCK=tck,
            o_TDI=self.tdi,
            o_SHIFT=shift,

            i_TDO=self.tdo,
        )
        return m



from typing import Iterable

from . import Stream, PacketizedStream
from nap.util.sim import wait_for

__all__ = ["write_to_stream", "read_from_stream", "read_packet_from_stream", "write_packet_to_stream"]


def write_to_stream(stream: Stream, timeout=100, **kwargs):
    for k, v in kwargs.items():
        yield stream[k].eq(v)
    yield stream.valid.eq(1)
    yield from wait_for(stream.ready, timeout)
    yield stream.valid.eq(0)


def read_from_stream(stream: Stream, extract="payload", timeout=100):
    yield stream.ready.eq(1)
    yield from wait_for(stream.valid, timeout)
    if isinstance(extract, str):
        read = (yield stream[extract])
    elif isinstance(extract, Iterable):
        read = []
        for x in extract:
            read.append((yield stream[x]))
        read = tuple(read)
    else:
        raise TypeError("extract must be either a string or an iterable of strings")

    yield stream.ready.eq(0)
    return read


def write_packet_to_stream(stream: PacketizedStream, payload_array, timeout=100):
    for i, p in enumerate(payload_array):
        if i < (len(payload_array) - 1):
            yield from write_to_stream(stream, timeout, payload=p, last=0)
        else:
            yield from write_to_stream(stream, timeout, payload=p, last=1)


def read_packet_from_stream(stream: PacketizedStream, timeout=100):
    packet = []
    while True:
        payload, last = yield from read_from_stream(stream, extract=("payload", "last"), timeout=timeout)
        packet.append(payload)
        if last:
            return packet



from nmigen import *
from ..generic.jtag import JTAG as GenericJTAG


class JTAG(GenericJTAG.implementation):
    # the lattice jtag primitive is rather wired and has a one cycle delay on tdi

    def elaborate(self, platform):
        m = Module()

        print("warning: lattice JTAGF primitive is buggy as fuck")

        cd = ClockDomain(self.jtag_domain, clk_edge="neg")  # we must sample tdo at the falling edge for some reason
        m.domains += cd
        platform.add_clock_constraint(cd.clk, 1e6)

        shift = Signal()  # for some reason we also have to delay shift
        m.d.comb += self.shift_tdo.eq(shift)
        m.d.jtag += self.shift_tdi.eq(shift)

        m.submodules.jtag_primitive = Instance(
            "JTAGF",
            i_JTDO1=self.tdo,
            o_JTDI=self.tdi,
            o_JTCK=cd.clk,
            o_JSHIFT=shift,
        )

        return m


import unittest
from nap.soc.pydriver.hardware_proxy import BitwiseAccessibleInteger


class BitwiseAccessibleIntegerTest(unittest.TestCase):
    def test_destruct(self):
        v = BitwiseAccessibleInteger(0b10100100)
        self.assertEqual(v[0], 0)
        self.assertEqual(v[2], 1)
        self.assertEqual(v[4:8], 0b1010)

    def test_construct(self):
        v = BitwiseAccessibleInteger(256)
        v[0] = 1
        self.assertEqual(int(v), 257)

        v = BitwiseAccessibleInteger()
        v[8] = 1
        self.assertEqual(int(v), 256)

        v = BitwiseAccessibleInteger(0b00001111)
        v[0:4] = 0
        self.assertEqual(int(v), 0)

        v = BitwiseAccessibleInteger(0b01011111)
        v[4:8] = 0b1010
        self.assertEqual(int(v), 0b10101111)


# An experiment that outputs raw data from dram with a mipi dsi phy.
from nmigen import *
from nmigen.build import Subsignal, Attrs, Resource, DiffPairs
from nap import *
from nap.vendor.xilinx_s7 import Pll


class Top(Elaboratable):
    def __init__(self):
        self.buffer_len = ControlSignal(32, reset=0x1000)
        self.buffer_base = ControlSignal(32, reset=0x0f80_0000)
        self.buffers_written = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        platform.ps7.fck_domain(100e6, "sync")

        platform.ps7.fck_domain(20e6, "pxclk")
        pll = m.submodules.pll = Pll(20e6, 60, 1, input_domain="pxclk")
        pll.output_domain("ddr_pxclk", 30)

        m.submodules.clocking = ClockingDebug("sync", "pxclk", "ddr_pxclk")

        # this experiment requires bodging the MicroR2 in a way that connects the southern plugin module
        # to a mipi DSI screen. Power and backlight have to be handled separated.
        platform.add_resources([
            Resource("mipi_dsi", 0,
                     Subsignal("clk", DiffPairs("25", "26", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("lane0", DiffPairs("27", "28", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
            ),
        ])
        panel = platform.request("mipi_dsi")

        is_initialized = Signal()
        address_stream = BasicStream(32)
        buffer_counter = Signal(32)

        m.d.comb += address_stream.valid.eq(is_initialized)
        with m.If(~is_initialized):
            m.d.sync += address_stream.payload.eq(self.buffer_base)
            m.d.sync += is_initialized.eq(1)
        with m.Elif(address_stream.ready):
            with m.If(buffer_counter < self.buffer_len):
                m.d.sync += address_stream.payload.eq(address_stream.payload + 8)
                m.d.sync += buffer_counter.eq(buffer_counter + 1)
            with m.Else():
                m.d.sync += address_stream.payload.eq(self.buffer_base)
                m.d.sync += buffer_counter.eq(0)
                m.d.sync += self.buffers_written.eq(self.buffers_written + 1)

        p = Pipeline(m)
        p += StreamBuffer(address_stream)
        p += AxiReader(p.output)
        p += StreamGearbox(p.output, target_width=8)
        p += BufferedAsyncStreamFIFO(p.output, depth=2048, o_domain="pxclk")
        p += MipiMultiLaneTxPhy(p.output, panel.clk, [panel.lane0], ddr_domain="ddr_pxclk")

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform,), possible_socs=(ZynqSocPlatform,))



# An experiment that glues everything together and tries to get a full sensor -> ethernet flow working on the micro
import os
from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        self.sensor_reset_n = ControlSignal(name='sensor_reset', reset=1)

    def elaborate(self, platform):
        m = Module()

        platform.ps7.fck_domain(100e6, "axi_hp")

        # Control Pane
        i2c_pads = platform.request("i2c")
        m.submodules.i2c = BitbangI2c(i2c_pads)

        # Input Pipeline
        sensor = platform.request("sensor")
        platform.ps7.fck_domain(24e6, "sensor_clk")
        m.d.comb += sensor.clk.eq(ClockSignal("sensor_clk"))
        m.d.comb += sensor.reset.eq(~self.sensor_reset_n)
        # TODO: find more idiomatic way to do this
        os.environ["NMIGEN_add_constraints"] = \
            "set_property CLOCK_DEDICATED_ROUTE FALSE [get_nets pin_sensor_0__lvds_clk/hispi_sensor_0__lvds_clk__i]"

        p = Pipeline(m)
        p += HispiRx(sensor, hispi_domain="hispi")
        p += Repack12BitStream(p.output)

        p += BufferedAsyncStreamFIFO(p.output, 2048, o_domain="axi_hp")
        p += StreamGearbox(p.output, 64)
        p += ImageStream2PacketizedStream(p.output)
        p += DramPacketRingbufferStreamWriter(p.output, max_packet_size=0x800000, n_buffers=4)
        p += DramPacketRingbufferCpuReader(p.last)

        return m

    @driver_method
    def kick_sensor(self):
        from os import system
        system("cat /axiom-api/scripts/kick/value")


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform,), possible_socs=(ZynqSocPlatform,))



import rawpy
import numpy as np
from pydng.core import RAW2DNG, DNGTags, Tag
from .py_wavelet import ty

positions = [(0, 0), (0, 1), (1, 0), (1, 1)]


def read_dng(filename):
    image = rawpy.imread(filename)
    raw_image = np.array(image.raw_image, dtype=ty)
    arrays = [raw_image[y::2, x::2] for x, y in positions]
    color_desc = image.color_desc.decode("utf-8")
    g = 1
    colors = {}
    for i, (x, y) in enumerate(positions):
        c = color_desc[image.raw_color(x, y)]
        if c == "G":
            c += str(g)
            g += 1
        colors[c] = arrays[i]

    estimated_bit_depth = 12 if np.max(raw_image) > 255 else 8
    return colors['R'], colors['G1'], colors['G2'], colors['B'], list(colors.keys()), estimated_bit_depth


def write_dng(filename, red, green1, green2, blue, bit_depth, order=('G1', 'R', 'B', 'G2')):
    h, w = red.shape
    result = np.empty((h * 2, w * 2), dtype=np.uint16)

    colors = {'R': red, 'G1': green1, 'G2': green2, 'B': blue}
    for (x, y), color in zip(positions, order):
        result[y::2, x::2] = colors[color]

    # set DNG tags.
    t = DNGTags()

    t.set(Tag.ImageWidth, w * 2)
    t.set(Tag.ImageLength, h * 2)
    t.set(Tag.TileWidth, w * 2)
    t.set(Tag.TileLength, h * 2)
    t.set(Tag.PhotometricInterpretation, 32803)
    t.set(Tag.SamplesPerPixel, 1)
    t.set(Tag.BitsPerSample, bit_depth)
    t.set(Tag.CFARepeatPatternDim, [2, 2])
    t.set(Tag.CFAPattern, [{'R': 0, 'G': 1, 'B': 2}[c[0]] for c in order])
    t.set(Tag.BlackLevel, 0)
    t.set(Tag.WhiteLevel, ((1 << bit_depth) - 1))
    t.set(Tag.CalibrationIlluminant1, 21)
    t.set(Tag.AsShotNeutral, [[1, 1], [1, 1], [1, 1]])
    t.set(Tag.DNGVersion, [1, 4, 0, 0])
    t.set(Tag.DNGBackwardVersion, [1, 2, 0, 0])
    t.set(Tag.Make, "Camera Brand")
    t.set(Tag.Model, "Camera Model")
    t.set(Tag.PreviewColorSpace, 2)

    # save to dng file.
    RAW2DNG().convert(result, tags=t, filename=filename)
    return filename + '.dng'



import huffman
from nmigen import *
from nap import PacketizedStream, stream_transformer
from . import VariableWidthStream

__all__ = ["HuffmanEncoder"]


class HuffmanEncoder(Elaboratable):
    def __init__(self, input: PacketizedStream, distribution):
        self.input = input

        self.distribution = distribution
        self.table = {k: v[::-1] for k, v in  huffman.codebook(self.distribution.items()).items()}

        self.max_input_word = max(self.distribution.keys()) + 1
        self.max_code_len = max([len(v) for v in self.table.values()])

        self.output = VariableWidthStream(self.max_code_len + 1)

    def elaborate(self, platform):
        m = Module()

        stream_transformer(self.input, self.output, m, latency=1, allow_partial_out_of_band=True)
        input_transaction = self.input.ready & self.input.valid

        code_memory = Memory(width=self.max_code_len, depth=self.max_input_word, init=[int(self.table.get(i, '0'), 2) for i in range(self.max_input_word)])
        code_port = m.submodules.code_port = code_memory.read_port(domain="sync", transparent=False)
        m.d.comb += code_port.en.eq(input_transaction)
        m.d.comb += code_port.addr.eq(self.input.payload)
        m.d.comb += self.output.payload.eq(code_port.data)

        code_len_memory = Memory(width=self.max_code_len, depth=self.max_input_word, init=[len(self.table.get(i, '')) for i in range(self.max_input_word)])
        len_port = m.submodules.len_port = code_len_memory.read_port(domain="sync", transparent=False)
        m.d.comb += len_port.en.eq(input_transaction)
        m.d.comb += len_port.addr.eq(self.input.payload)
        m.d.comb += self.output.current_width.eq(len_port.data)

        return m



import unittest
from nmigen.build import Clock
from nap import SimPlatform
from . import generate_modeline, parse_modeline, HdmiClocking, HdmiTimingGenerator


class TestHdmi(unittest.TestCase):
    def test_timing_generator(self):
        platform = SimPlatform()
        dut = HdmiTimingGenerator(parse_modeline(generate_modeline(640, 480, 60)))

        def testbench():
            last_x = 0
            for i in range(800 - 1):
                yield
                this_x = (yield dut.x)
                assert this_x == last_x + 1, "x increment failed"
                last_x = this_x
            yield
            assert 1 == (yield dut.y), "y increment failed"

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut, testbench)

    def test_mmcm_calculation(self):
        clocking = HdmiClocking(Clock(79.75e6), pix_domain="pix")
        clocking.find_valid_config()



from nmigen import Signal
from nap import BasicStream, DOWNWARDS

__all__ = ["ImageStream"]


class ImageStream(BasicStream):
    """
    A stream that can be used to transfer image data.
    """

    def __init__(self, payload_shape, name=None, src_loc_at=1):
        super().__init__(payload_shape, name, src_loc_at=1 + src_loc_at)
        self.line_last = Signal() @ DOWNWARDS
        self.frame_last = Signal() @ DOWNWARDS



from nmigen import *
from nap import ControlSignal, StatusSignal, driver_method
from .stream_if import DramPacketRingbufferStreamWriter

__all__ = ["DramPacketRingbufferCpuReader", "DramPacketRingbufferCpuWriter"]


class DramPacketRingbufferCpuWriter(Elaboratable):
    def __init__(
            self,
            max_packet_size, n_buffers, base_address=0x0f80_0000, default_packet_size=0
    ):
        self.max_packet_size = max_packet_size
        self.base_address = base_address
        self.n_buffers = n_buffers

        self.buffer_base_list = Array([base_address + max_packet_size * i for i in range(n_buffers)])
        self.buffer_level_list = Array([ControlSignal(range(max_packet_size), reset=default_packet_size) for _ in range(n_buffers)])
        self.current_write_buffer = ControlSignal(range(n_buffers))

        for i, signal in enumerate(self.buffer_level_list):
            setattr(self, f"buffer{i}_level", signal)

    def elaborate(self, platform):
        return Module()

    # TODO: implement driver methods


class DramPacketRingbufferCpuReader(Elaboratable):
    def __init__(self, writer: DramPacketRingbufferStreamWriter):
        self.writer = writer

        m = Module()

        self.current_write_buffer = StatusSignal(range(writer.n_buffers))
        m.d.comb += self.current_write_buffer.eq(writer.current_write_buffer)

        for i, (base, level) in enumerate(zip(writer.buffer_base_list, writer.buffer_level_list)):
            print(base, level)
            buffer_base = StatusSignal(range(base + 1))
            m.d.comb += buffer_base.eq(base)
            setattr(self, f"buffer{i}_base", buffer_base)

            buffer_level = StatusSignal(level.shape())
            m.d.comb += buffer_level.eq(level)
            setattr(self, f"buffer{i}_level", buffer_level)

        self.m = m

    def elaborate(self, platform):
        return self.m

    @driver_method
    def read_packet_to_file(self, filename="packet.bin"):
        from time import sleep
        import os
        while not self.next_buffer_ready:
            sleep(0.01)

        self.current_read_buffer = self.current_read_buffer % self.num_buffers
        os.system("dd if=/dev/mem bs=4096 skip={} count={} iflag=skip_bytes,count_bytes of='{}'"
                  .format(self.current_buffer_base, self.current_buffer_length, filename))




from nmigen import *
from nap.soc import ControlSignal, StatusSignal, devicetree_overlay

__all__ = ["MmioGpio"]


class MmioGpio(Elaboratable):
    def __init__(self, pads, name_suffix=""):
        """ A simple gpio peripheral, that is compatible with the gpio-mmio.c linux kernel pydriver.
        see https://github.com/torvalds/linux/blob/master/drivers/gpio/gpio-mmio.c
        """
        self._pads = pads

        # see https://github.com/torvalds/linux/blob/master/drivers/gpio/gpio-mmio.c#L473
        # we are using a configuration with one output one input and one direction register
        w = len(self._pads)
        self.set = ControlSignal(w)
        self.dat = StatusSignal(w)
        self.dirout = ControlSignal(w)

        self.devicetree_name = "mmio_gpio" + name_suffix

    def elaborate(self, platform):
        m = Module()

        overlay_content = """
            %overlay_name%: %overlay_name%@40000000 {
                    compatible = "brcm,bcm6345-gpio";
                    reg-names = "set", "dat", "dirout";
                    reg = <%set% 1>, <%dat% 1>, <%dirout% 1>;
            
                    #gpio-cells = <2>;
                    gpio-controller;
            };
        """
        devicetree_overlay(platform, self.devicetree_name, overlay_content, {"set": self.set, "dat": self.dat, "dirout": self.dirout})

        for i, pad in enumerate(self._pads):
            m.d.comb += self.dat[i].eq(pad.i)
            m.d.comb += pad.oe.eq(self.dirout[i])
            m.d.comb += pad.o.eq(self.set[i])

        return m



class JTAGAccessor:
    base = 0

    def __init__(self, addr="127.0.0.1", port=4444, timeout=1024, debug=False, spawn_server=True, tap_name="dut.tap"):
        import socket
        import os
        import time
        from os.path import dirname, abspath

        self.tap_name = tap_name

        if spawn_server:
            addr = "127.0.0.1"
            port = 4444
            os.system('cd {}; openocd -f openocd.cfg > /dev/null 2>&1 &'.format(dirname(abspath(__file__))))
            time.sleep(0.1)

        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.connect((addr, port))
        self.timeout = timeout
        self.debug = debug
        self.spawn_server = spawn_server

        # skip some strange random shit
        self._readline(decode=False)
        self._irscan(0x32)
        for i in range(3):
            self._shift_word(0)

    def __del__(self):
        if hasattr(self, "spawn_server") and self.spawn_server:
            self._writeline("shutdown")

    def read(self, addr):
        self._shift_bit(0)  # wakeup
        self._shift_bit(1)  # wakeup
        self._shift_word(addr)  # address
        self._shift_bit(0)  # read

        # read wait
        timeout = self.timeout
        for t in range(timeout):
            if self._shift_bit(1) == 1:
                break
            if t == timeout - 1:
                raise TimeoutError()

        data = self._shift_word(0)
        if self._shift_bit(0) != 0:  # read status
            raise TransactionNotSuccessfulException()

        return data

    def write(self, addr, value):
        self._shift_bit(0)  # wakeup
        self._shift_bit(1)  # wakeup
        self._shift_word(addr)  # address
        self._shift_bit(1)  # write
        self._shift_word(value)

        # write wait
        timeout = self.timeout
        for t in range(timeout):
            if self._shift_bit(1) == 1:
                break
            if t == timeout - 1:
                raise TimeoutError()
        if self._shift_bit(0) != 0:  # write status
            raise TransactionNotSuccessfulException()

    def _writeline(self, message):
        message += "\n"
        self.s.send(message.encode('utf-8'))

    def _readline(self, decode=True):
        buf = b""

        while buf[-2:] != b'> ':
            buf += self.s.recv(1)

        if decode and self.debug:
            print(buf.decode('utf-8'))

        buf = buf.rsplit(b'\r\n', 2)[1][1:]

        if decode:
            buf = buf.decode('utf-8')

        return buf

    def _writecmd(self, cmd):
        self._writeline(cmd)
        ret = self._readline()
        if self.debug:
            print(cmd, "->", ret)
        return ret

    def _shift_word(self, write):
        return int(self._drscan(32, write), 16)

    def _shift_bit(self, write):
        return int(self._drscan(1, write), 16)

    def _irscan(self, instruction):
        return self._writecmd('irscan {} {}'.format(self.tap_name, instruction))

    def _drscan(self, len, value):
        return self._writecmd('drscan {} {} {}'.format(self.tap_name, len, value))


MemoryAccessor = JTAGAccessor


class TransactionNotSuccessfulException(Exception):
    pass



from nmigen import *
from nap import StatusSignal, BasicStream, PacketizedStream
from nap.cores import StreamInfo, BufferedSyncStreamFIFO, StreamTee
from .axi_endpoint import AxiAddressStream, AxiDataStream, AxiResponse
from .stream_reader import AxiReaderBurster
from .zynq_util import if_none_get_zynq_hp_port

__all__ = ["AxiWriter"]


class AxiWriter(Elaboratable):
    def __init__(
            self,
            address_source: BasicStream,
            data_source: BasicStream,
            axi=None
    ):
        self.address_source = address_source
        self.data_source = data_source
        self.axi = axi

        self.axi_address_ready = StatusSignal()
        self.axi_data_ready = StatusSignal()
        self.write_response_ok = StatusSignal(32)
        self.write_response_err = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        burster = m.submodules.burster = AxiWriterBurster(self.address_source, self.data_source)

        axi = if_none_get_zynq_hp_port(self.axi, m, platform)
        for fifo_signal_name in ["write_address_fifo_level", "write_data_fifo_level"]:
            if hasattr(axi, fifo_signal_name):
                axi_fifo_signal = axi[fifo_signal_name]
                fifo_signal = StatusSignal(axi_fifo_signal.shape(), name=f"axi_{fifo_signal_name}")
                m.d.comb += fifo_signal.eq(axi_fifo_signal)
                setattr(self, f"axi_{fifo_signal_name}", fifo_signal)

        m.d.comb += axi.write_data.connect_upstream(burster.data_output)
        m.d.comb += axi.write_address.connect_upstream(burster.address_output)

        # we do not currently care about the write responses
        m.d.comb += axi.write_response.ready.eq(1)
        with m.If(axi.write_response.valid):
            with m.If(axi.write_response.resp == AxiResponse.OKAY):
                m.d.sync += self.write_response_ok.eq(self.write_response_ok + 1)
            with m.Else():
                m.d.sync += self.write_response_err.eq(self.write_response_err + 1)

        m.d.comb += self.axi_data_ready.eq(axi.write_data.ready)
        m.d.comb += self.axi_address_ready.eq(axi.write_address.ready)

        info_axi_address = m.submodules.info_axi_address = StreamInfo(axi.write_address)
        info_axi_data = m.submodules.info_axi_data = StreamInfo(axi.write_data)

        return m


class AxiWriterBurster(Elaboratable):
    """Creates Write bursts. Outputs Burst addresses on the address Channel and modifies the Data channels last accordingly."""

    def __init__(
            self,
            address_source: BasicStream,
            data_source: BasicStream,
            max_burst_length=16, burst_creation_timeout=31,
            data_fifo_depth=16,
    ):
        self.max_burst_length = max_burst_length
        self.burst_creation_timeout = burst_creation_timeout
        self.word_bytes = len(data_source.payload) // 8
        self.data_fifo_depth = data_fifo_depth

        self.address_input = address_source
        self.data_input = data_source

        self.written_address_bursts_for_n_wards = StatusSignal(32)

        self.address_output = AxiAddressStream(addr_bits=len(address_source.payload), lite=False, id_bits=12, data_bytes=self.word_bytes)
        self.data_output = AxiDataStream(data_bits=len(data_source.payload), read=False, lite=False, id_bits=12)

    def elaborate(self, platform):
        m = Module()
        address_burster = m.submodules.address_burster = AxiReaderBurster(self.address_input, self.word_bytes, self.max_burst_length, self.burst_creation_timeout)
        burst_tee = m.submodules.burst_tee = StreamTee(address_burster.output)
        m.d.comb += self.address_output.connect_upstream(burst_tee.get_output())

        burst_stream = burst_tee.get_output()
        packet_length_stream = BasicStream(burst_stream.burst_len.shape())
        m.d.comb += packet_length_stream.valid.eq(burst_stream.valid)
        m.d.comb += burst_stream.ready.eq(packet_length_stream.ready)
        m.d.comb += packet_length_stream.payload.eq(burst_stream.burst_len)

        data_fifo = m.submodules.data_fifo = BufferedSyncStreamFIFO(self.data_input, self.data_fifo_depth)
        stream_packetizer = m.submodules.stream_packetizer = StreamPacketizer(packet_length_stream, data_fifo.output)
        m.d.comb += self.data_output.connect_upstream(stream_packetizer.output, allow_partial=True)
        m.d.comb += self.data_output.id.eq(self.data_output.id.reset)
        m.d.comb += self.data_output.byte_strobe.eq(-1)

        return m


class StreamPacketizer(Elaboratable):
    def __init__(self, packet_length_stream: BasicStream, data_stream: BasicStream):
        self.packet_length_stream = packet_length_stream
        self.data_stream = data_stream

        self.output = PacketizedStream(self.data_stream.payload.shape())

    def elaborate(self, platform):
        m = Module()
        m.d.comb += self.output.payload.eq(self.data_stream.payload)

        packet_counter = Signal(self.packet_length_stream.payload.shape())
        with m.If(self.packet_length_stream.valid & self.data_stream.valid):
            m.d.comb += self.output.valid.eq(1)
            with m.If(self.output.ready):
                m.d.comb += self.data_stream.ready.eq(1)

            with m.If(packet_counter < self.packet_length_stream.payload):
                with m.If(self.output.ready):
                    m.d.sync += packet_counter.eq(packet_counter + 1)
            with m.Else():
                m.d.comb += self.output.last.eq(1)
                with m.If(self.output.ready):
                    m.d.comb += self.packet_length_stream.ready.eq(1)
                    m.d.sync += packet_counter.eq(0)

        return m



from nmigen import *
from nap import BasicStream, StatusSignal, ControlSignal, driver_method
from nap.cores import InflexibleSourceDebug
from nap.vendor.lattice_machxo2 import ISerdes8, Pll, EClkSync, ClkDiv

__all__ = ["PluginModuleStreamerRx"]


class PluginModuleStreamerRx(Elaboratable):
    def __init__(self, plugin, domain_name="sync"):
        self.plugin = plugin
        self.output = BasicStream(32)
        self.domain_name = domain_name

        self.trained = ControlSignal()
        self.valid = StatusSignal()

    def elaborate(self, platform):
        m = Module()

        domain = self.domain_name
        domain_in = domain + "_in"
        domain_ddr = domain + "_ddr"
        domain_ddr_eclk = domain + "_ddr_eclk"

        m.domains += ClockDomain(domain_in)
        m.d.comb += ClockSignal(domain_in).eq(self.plugin.clk_word)

        pll = m.submodules.pll = Pll(input_freq=50e6, vco_mul=4, vco_div=1, input_domain=domain_in)
        pll.output_domain(domain_ddr, 1)

        m.submodules.eclk_ddr = EClkSync(domain_ddr, domain_ddr_eclk, input_frequency=400e6)

        lanes = []
        bitslip_signal = Signal()
        for i in range(4):
            lane = m.submodules["lane{}".format(i)] = LaneBitAligner(
                input=self.plugin["lvds{}".format(i)],
                in_testpattern_mode=~self.valid,
                bitslip_signal=bitslip_signal,
                ddr_domain=domain_ddr_eclk,
            )
            lanes.append(lane)
        word_aligner = m.submodules.word_aligner = WordAligner(domain_ddr_eclk, lanes[0].output)
        m.d.comb += bitslip_signal.eq(word_aligner.bitslip)

        valid_iserdes = m.submodules.valid_iserdes = ISerdes8(self.plugin.valid, domain_ddr_eclk, word_domain="sync", invert=True)
        m.d.comb += valid_iserdes.bitslip.eq(bitslip_signal)
        m.d.comb += self.valid.eq(valid_iserdes.output[4])

        m.d.sync += self.output.payload.eq(Cat(*[lane.output for lane in lanes]))
        m.d.comb += self.output.valid.eq(
            self.valid & self.trained
        )

        m.submodules.inflexible_output = InflexibleSourceDebug(self.output)

        return DomainRenamer(domain)(m)

    @driver_method
    def train(self, timeout=32):
        self.lane0.delay = 15
        print("doing word alignment...")
        for i in range(timeout):
            if self.lane0.output == 0b00010110:
                print("-> {} slips".format(i))
                print("training lane 0...")
                self.lane0.train()
                print("training lane 1...")
                self.lane1.train()
                print("training lane 2...")
                self.lane2.train()
                print("training lane 3...")
                self.lane3.train()
                self.trained = True
                return
            else:
                self.word_aligner.slip()
        raise TimeoutError()


class WordAligner(Elaboratable):
    def __init__(self, ddr_domain, lane_output):
        self.ddr_domain = ddr_domain

        self.lane_output = lane_output
        self.do_bitslip = ControlSignal()

        self.bitslip = Signal()

    def elaborate(self, platform):
        m = Module()

        m.submodules.clk_div_quater = ClkDiv(self.ddr_domain, "sync", div=4, input_frequency=400e6, bitslip=self.bitslip)

        last_do_bitslip = Signal()
        m.d.sync += last_do_bitslip.eq(self.do_bitslip)
        with m.If(last_do_bitslip != self.do_bitslip):
            m.d.comb += self.bitslip.eq(1)

        return m

    @driver_method
    def slip(self):
        self.do_bitslip = not self.do_bitslip


class LaneBitAligner(Elaboratable):
    def __init__(self, input, in_testpattern_mode, ddr_domain, bitslip_signal, testpattern=0b00010110):
        """
        Does bit alignment of one lane usig a given testpattern if in_testpattern_mode is high
        """
        self.input = input
        self.bitslip_signal = bitslip_signal
        self.in_testpattern_mode = in_testpattern_mode

        self.ddr_domain = ddr_domain
        self.testpattern = testpattern

        self.delay = ControlSignal(5, reset=15)
        self.error = StatusSignal(32)

        self.output = StatusSignal(8)

    def elaborate(self, platform):
        m = Module()

        delayed = Signal()
        m.submodules.delayd = Instance(
            "DELAYD",

            i_A=self.input,
            i_DEL0=self.delay[0],
            i_DEL1=self.delay[1],
            i_DEL2=self.delay[2],
            i_DEL3=self.delay[3],
            i_DEL4=self.delay[4],

            o_Z=delayed,
        )
        iserdes = m.submodules.iserdes = ISerdes8(delayed, self.ddr_domain, word_domain="sync", invert=True)
        m.d.comb += self.output.eq(iserdes.output)
        m.d.comb += iserdes.bitslip.eq(self.bitslip_signal)

        with m.If(self.in_testpattern_mode & (self.output != self.testpattern)):
            m.d.sync += self.error.eq(self.error + 1)

        return m

    @driver_method
    def train(self):
        from time import sleep
        start_current = 0
        start_longest = 0
        len_longest = 0
        was_good = True
        for i in range(32):
            self.delay = i
            e_start = self.error
            sleep(0.01)
            difference = self.error - e_start
            if difference == 0 and not was_good:
                start_current = i
            elif ((difference != 0) or (i == 31)) and was_good:
                len = i - start_current
                if len > len_longest:
                    len_longest = len
                    start_longest = start_current
            print(i, difference)
            was_good = difference == 0
        self.delay = int(start_longest + (len_longest / 2))
        print("-> delay tap", self.delay)



import copy
from collections import OrderedDict
from nmigen import *
from nap import Bundle, UPWARDS, DOWNWARDS

__all__ = ["Stream", "BasicStream", "PacketizedStream"]


class Stream(Bundle):
    """
    A stream is everything that inherits from `Bundle` and has a ready@UPWARDS, a valid@DOWNWARDS.
    Stream Sinks (downstream) pull ready high if they can accept data and stream Sources (upstream) pull valid
    high if they have data to offer. Both ready-before-valid and valid-before-ready are allowed.
    A successful data transfer on the stream happens if ready and valid are 1 during a cycle.

    Optionally Streams may have any number of DOWNWARDS Signals, that carry actual payload and/or out-of-band signaling
    (e.g. last or first) as their payload.

    Streams have to implement a clone() methods that returns a Stream that is shaped just like the original stream
    but not connected to it.

    This is the base Stream class that is pretty much useless on its own. Other Classes are implementing Streams that
    are useful for real use-cases. Probably you are looking for `BasicStream`
    """

    def __init__(self, name=None, src_loc_at=1):
        super().__init__(name=name, src_loc_at=1 + src_loc_at)

        self.ready = Signal() @ UPWARDS
        self.valid = Signal() @ DOWNWARDS

    def clone(self, name=None, src_loc_at=1):
        new_stream = self.__class__.__new__(self.__class__)
        Stream.__init__(new_stream, name=name, src_loc_at=1 + src_loc_at)

        for k, signal in self.payload_signals.items():
            setattr(new_stream, k, Signal(signal.shape()) @ DOWNWARDS)
        new_stream._directions = copy.deepcopy(self._directions)
        return new_stream

    @property
    def payload_signals(self):
        return OrderedDict((k, self[k]) for k, direction in self._directions.items() if k != "valid" and direction == DOWNWARDS)


class BasicStream(Stream):
    """A basic stream that carries a payload"""

    def __init__(self, payload_shape, name=None, src_loc_at=1):
        super().__init__(name=name, src_loc_at=1 + src_loc_at)
        self.payload = Signal(payload_shape) @ DOWNWARDS

    @property
    def out_of_band_signals(self):
        return OrderedDict((k, v) for k, v in self.payload_signals.items() if k != "payload")


class PacketizedStream(BasicStream):
    """
    A stream that carries a payload and can separate Packets via a last signal that is asserted on the
    last word of a packet
    """

    def __init__(self, payload_shape, name=None, src_loc_at=1):
        super().__init__(payload_shape, name, src_loc_at=1 + src_loc_at)
        self.last = Signal() @ DOWNWARDS



from functools import reduce

from nmigen import *
from nmigen.hdl.ast import SignalSet

from .csr_types import _Csr, ControlSignal, StatusSignal, EventReg
from .memorymap import MemoryMap
from .pydriver.drivermethod import DriverMethod
from .tracing_elaborate import ElaboratableSames


def csr_and_driver_method_hook(platform, top_fragment: Fragment, sames: ElaboratableSames):
    from nap.cores.peripherals.csr_bank import CsrBank
    already_done = []

    def inner(fragment):
        elaboratable = sames.get_elaboratable(fragment)
        if elaboratable:
            class_members = [(s, getattr(elaboratable, s)) for s in dir(elaboratable)]
            csr_signals = [(name, member) for name, member in class_members if isinstance(member, _Csr)]
            fragment_signals = reduce(lambda a, b: a | b, fragment.drivers.values(), SignalSet())
            csr_signals += [
                (signal.name, signal) for signal in fragment_signals
                if isinstance(signal, _Csr)
                   and signal.name != "$signal"
                   and not any(signal is cmp_signal for name, cmp_signal in csr_signals)
            ]

            new_csr_signals = [(name, signal) for name, signal in csr_signals if not any(signal is done for done in already_done)]
            old_csr_signals = [(name, signal) for name, signal in csr_signals if any(signal is done for done in already_done)]
            for name, signal in new_csr_signals:
                already_done.append(signal)

            mmap = fragment.memorymap = MemoryMap()

            if new_csr_signals:
                m = Module()
                csr_bank = m.submodules.csr_bank = CsrBank()
                for name, signal in new_csr_signals:
                    if isinstance(signal, (ControlSignal, StatusSignal, EventReg)):
                        csr_bank.reg(name, signal)
                        signal._MustUse__used = True

                mmap.allocate_subrange(csr_bank.memorymap, name=None)  # name=None means that the Memorymap will be inlined
                platform.to_inject_subfragments.append((m, "ignore"))

            for name, signal in old_csr_signals:
                mmap.add_alias(name, signal)

            driver_methods = [(name, getattr(elaboratable, name)) for name in dir(elaboratable)
                              if isinstance(getattr(elaboratable, name), DriverMethod)]
            for name, driver_method in driver_methods:
                fragment.memorymap.add_driver_method(name, driver_method)

        for subfragment, name in fragment.subfragments:
            inner(subfragment)
    inner(top_fragment)


def address_assignment_hook(platform, top_fragment: Fragment, sames: ElaboratableSames):
    def inner(fragment):
        module = sames.get_module(fragment)
        if hasattr(module, "peripheral"):  # we have the fragment of a marker module for a peripheral
            fragment.memorymap = module.peripheral.memorymap
            return

        # depth first recursion is important so that all the subfragments have a fully populated memorymap later on
        for sub_fragment, sub_name in fragment.subfragments:
            if sub_name != "ignore":
                inner(sub_fragment)

        # add everything to the own memorymap
        if not hasattr(fragment, "memorymap"):
            fragment.memorymap = MemoryMap()
        for sub_fragment, sub_name in fragment.subfragments:
            if sub_name != "ignore":
                assert hasattr(sub_fragment, "memorymap")  # this holds because we did depth first recursion
                fragment.memorymap.allocate_subrange(sub_fragment.memorymap, sub_name)
    inner(top_fragment)

    # prepare and finalize the memorymap
    top_memorymap: MemoryMap = top_fragment.memorymap
    top_memorymap.is_top = True

    assert platform.base_address is not None
    top_memorymap.place_at = platform.base_address

    print("memorymap:\n" + "\n".join(
        "    {}: {!r}".format(".".join(k), v) for k, v in top_memorymap.flattened.items()))
    platform.memorymap = top_memorymap


def peripherals_collect_hook(platform, top_fragment: Fragment, sames: ElaboratableSames):
    platform.peripherals = []

    def collect_peripherals(platform, fragment: Fragment, sames):
        module = sames.get_module(fragment)
        if module:
            if hasattr(module, "peripheral"):
                platform.peripherals.append(module.peripheral)
        for (f, name) in fragment.subfragments:
            collect_peripherals(platform, f, sames)

    collect_peripherals(platform, top_fragment, sames)

    ranges = [(peripheral.range(), peripheral) for peripheral in platform.peripherals
              if not peripheral.memorymap.is_empty and not peripheral.memorymap.was_inlined]

    def range_overlapping(x, y):
        if x.start == x.stop or y.start == y.stop:
            return False
        return ((x.start < y.stop and x.stop > y.start) or
                (x.stop > y.start and y.stop > x.start))

    for a, peripheral_a in ranges:
        for b, peripheral_b in ranges:
            if a is not b and range_overlapping(a, b):
                raise AssertionError("{!r} overlaps with {!r}".format(peripheral_a, peripheral_b))



from nmigen import *
from nap.stream import PacketizedStream
from .encoding_space import EncodingSpace

__all__ = ["RleEncodingSpace", "ZeroRleEncoder"]


class RleEncodingSpace(EncodingSpace):
    def __init__(self, input_range: range, possible_run_lengths_list, zero_value):
        self.zero_value = zero_value
        self.input_range = input_range
        self.possible_run_lengths_list = possible_run_lengths_list
        self.max_run_length = max(possible_run_lengths_list)

    @property
    def numeric_range(self):
        return range(self.input_range.start, self.input_range.stop + len(self.possible_run_lengths_list))


class ZeroRleEncoder(Elaboratable):
    """
    Converts a stream of numbers into a stream of numbers (with identity mapping) and numbers for different
    run lengths of zeroes (with a dict of possible run lengths).
    """

    def __init__(self, input: PacketizedStream, encoding_space: RleEncodingSpace):
        self.encoding_space = encoding_space
        self.input = input

        self.output = PacketizedStream(encoding_space.numeric_range)

    def elaborate(self, platform):
        m = Module()

        run_length = Signal(range(self.encoding_space.max_run_length))
        with m.FSM():
            with m.State("NONZERO"):
                with m.If(((self.input.payload != self.encoding_space.zero_value) | self.input.last) & self.input.valid):
                    m.d.comb += self.output.connect_upstream(self.input)
                    m.d.sync += run_length.eq(0)
                with m.Elif(self.input.valid):
                    m.next = "ZERO_COUNT"
                    m.d.comb += self.input.ready.eq(1)
                    m.d.sync += run_length.eq(1)

            run_length_index = Signal(range(len(self.encoding_space.possible_run_lengths_list)))
            run_length_array = Array([1, *self.encoding_space.possible_run_lengths_list])

            with m.State("ZERO_COUNT"):
                with m.If(((self.input.payload != self.encoding_space.zero_value) | self.input.last) & self.input.valid):
                    m.next = "ZERO_OUT"
                    m.d.sync += run_length_index.eq(len(self.encoding_space.possible_run_lengths_list))
                with m.Elif(self.input.valid & (run_length < self.encoding_space.max_run_length - 1)):
                    m.d.comb += self.input.ready.eq(1)
                    m.d.sync += run_length.eq(run_length + 1)
                with m.Elif(self.input.valid):
                    m.d.comb += self.input.ready.eq(1)
                    m.d.sync += run_length.eq(run_length + 1)
                    m.next = "ZERO_OUT"
                    m.d.sync += run_length_index.eq(len(self.encoding_space.possible_run_lengths_list))

            with m.State("ZERO_OUT"):
                with m.If(run_length == 0):
                    m.next = "NONZERO"
                with m.Elif((run_length_array[run_length_index] <= run_length) & self.output.ready):
                    with m.If(run_length_index == 0):
                        m.d.comb += self.output.payload.eq(self.encoding_space.zero_value)
                    with m.Else():
                        m.d.comb += self.output.payload.eq((2 ** len(self.input.payload)) + run_length_index - 1)
                    m.d.comb += self.output.valid.eq(1)
                    m.d.sync += run_length.eq(run_length - run_length_array[run_length_index])
                with m.Elif(self.output.ready):
                    m.d.sync += run_length_index.eq(run_length_index - 1)

        return m



from textwrap import dedent

from nmigen.build import *
from nmigen.vendor.lattice_machxo_2_3l import *

from nap import FatbitstreamContext, program_bitstream_ssh

__all__ = ["Usb3PluginPlatform"]


class Usb3PluginPlatform(LatticeMachXO2Platform):
    device = "LCMXO2-2000HC"
    package = "TG100"
    speed = "6"

    resources = [
        Resource("ft601", 0,
            Subsignal("reset", PinsN("4", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

            Subsignal("data", Pins("75 74 70 69 68 67 66 65 64 61 60 59 58 57 54 53 83 84 85 86 87 88 96 97 98 99 7 8 21 24 20 25", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("be", Pins("19 18 17 16", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("oe", PinsN("9", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

            Subsignal("read", PinsN("10", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("write", PinsN("12", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("siwu", PinsN("13", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("rxf", PinsN("14", dir="i"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("txe", PinsN("15", dir="i"), Attrs(IO_TYPE="LVCMOS33")),

            Subsignal("gpio", Pins("2 1", dir="io"), Attrs(IO_TYPE="LVCMOS33")),
            Subsignal("wakeup", PinsN("3", dir="io"), Attrs(IO_TYPE="LVCMOS33")),

            Subsignal("clk", Pins("63", dir="i"), Attrs(IO_TYPE="LVCMOS33"), Clock(100e6)),
            Subsignal("clk1", Pins("62", dir="i"), Attrs(IO_TYPE="LVCMOS33"), Clock(100e6)),

            ),

        Resource("led", 0, Pins("71", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

        Resource("plugin_stream_input", 0,
            # BEWARE: all of these inputs are inverted but cant be inverted here (they have to be inverted after the iserdes)
            Subsignal("valid", DiffPairs("45", "47", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100")),  # lvds0
            Subsignal("lvds0", DiffPairs("42", "43", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100")),  # lvds1
            Subsignal("lvds1", DiffPairs("40", "41", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100")),  # lvds2
            Subsignal("lvds2", DiffPairs("36", "37", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100")),  # lvds3 TODO: move to 37 / 38 in the next rev
            Subsignal("lvds3", DiffPairs("29", "30", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100")),  # lvds4
            Subsignal("clk_word", DiffPairs("34", "35", dir="i"), Attrs(IO_TYPE="LVDS25", DIFFRESISTOR="100"), Clock(50e6)),  # lvds5
        )
    ]
    connectors = []

    def __init__(self):
        super().__init__()
        fc = FatbitstreamContext.get(self)
        fc.self_extracting_blobs["openocd_micro.cfg"] = dedent(r"""
            adapter driver sysfsgpio
    
            sysfsgpio_tms_num 898
            sysfsgpio_tck_num 899
            sysfsgpio_tdi_num 900
            sysfsgpio_tdo_num 901
            
            bindto 0.0.0.0
            transport select jtag
            jtag newtap dut tap -expected-id 0x012bb043 -irlen 8 -irmask 0xFF
            init
            scan_chain
        """)

        fc.self_extracting_blobs["openocd_beta.cfg"] = dedent(r"""
            adapter driver rfdev_jtag
            rfdev_set_device /dev/rfjtag1 
    
            bindto 0.0.0.0
            transport select jtag
            jtag newtap dut tap -expected-id 0x012bb043 -irlen 8 -irmask 0xFF
            init
            scan_chain
        """)

        fc.pre_init_commands.append("[ $(cat /etc/hostname) == 'beta' ] && cp openocd_beta.cfg openocd.cfg || cp openocd_micro.cfg openocd.cfg")

    def toolchain_program(self, *args, **kwargs):
        program_bitstream_ssh(self, *args, **kwargs)



import math
from enum import Enum

from nmigen import *

from nap.stream import BasicStream, Stream
from nap.data_structure import packed_struct, Bundle, UPWARDS, DOWNWARDS

__all__ = ["AxiResponse", "AxiBurstType", "AxiProtectionType", "AxiAddressStream", "AxiDataStream", "AxiWriteResponseStream", "AxiEndpoint"]


class AxiResponse(Enum):
    OKAY = 0b00
    EXOKAY = 0b01
    SLVERR = 0b10
    DECERR = 0b11


class AxiBurstType(Enum):
    FIXED = 0b00
    INCR = 0b01
    WRAP = 0b10


@packed_struct
class AxiProtectionType:
    privileged: unsigned(1)
    secure: unsigned(1)
    is_instruction: unsigned(1)


class AxiAddressStream(BasicStream):
    def __init__(self, addr_bits, lite, id_bits, data_bytes, src_loc_at=1):
        assert addr_bits % 8 == 0
        super().__init__(addr_bits, src_loc_at=1 + src_loc_at)
        if not lite:
            self.id = Signal(id_bits)
            self.burst_type = Signal(AxiBurstType)
            self.burst_len = Signal(range(16))  # in axi3 a burst can have a length of 16 as a hardcoded maximum
            self.beat_size_bytes = Signal(3, reset=int(math.log2(data_bytes)))  # this is 2**n encoded
            self.protection_type = Signal(AxiProtectionType().as_value().shape())


class AxiDataStream(BasicStream):
    def __init__(self, data_bits, read, lite, id_bits, src_loc_at=1, **kwargs):
        assert data_bits % 8 == 0
        super().__init__(data_bits, src_loc_at=1 + src_loc_at, **kwargs)
        if read:
            self.resp = Signal(AxiResponse)
        else:
            self.byte_strobe = Signal(data_bits // 8)

        if not lite:
            self.last = Signal()
            self.id = Signal(id_bits)


class AxiWriteResponseStream(Stream):
    def __init__(self, lite, id_bits, src_loc_at=1, **kwargs):
        super().__init__(src_loc_at=1 + src_loc_at, **kwargs)

        self.ready = Signal() @ UPWARDS
        self.valid = Signal()
        self.resp = Signal(AxiResponse)

        if not lite:
            self.id = Signal(id_bits)


class AxiEndpoint(Bundle):
    @staticmethod
    def like(model, lite=None, name="axi", **kwargs):
        """
        Create an AxiInterface shaped like a given model.
        :param name: the name of the resulting axi port
        :type model: AxiEndpoint
        :param model: the model after which the axi port should be created
        :param lite: overrides the lite property of the model. Only works for creating an AXI lite inteface from an AXI full bus.
        :return:
        """
        if lite is False:
            assert not model.is_lite, "cant make up id_bits out of thin air"

        return AxiEndpoint(
            addr_bits=model.addr_bits,
            data_bits=model.data_bits,
            lite=lite if lite is not None else model.is_lite,
            id_bits=None if (lite is not None and lite) else model.id_bits,
            name=name,
            **kwargs
        )

    def __init__(self, *, addr_bits, data_bits, lite, id_bits=None, src_loc_at=1, **kwargs):
        """
        Constructs a Record holding all the signals for axi (lite)

        :param addr_bits: the number of bits the address signal has.
        :param data_bits: the number of bits the
        :param lite: whether to construct an AXI lite or full bus.
        :param id_bits: the number of bits for the id signal of full axi. do not specify if :param lite is True.
        """
        super().__init__(**kwargs, src_loc_at=1 + src_loc_at)

        if id_bits:
            assert not lite, "there is no id tracking on axi lite buses"
        if lite:
            assert id_bits is None
        self.addr_bits = addr_bits
        self.data_bits = data_bits
        self.data_bytes = data_bits // 8
        assert math.log2(self.data_bytes) == int(math.log2(self.data_bytes))
        self.is_lite = lite
        self.id_bits = id_bits

        lite_args = {"lite": lite, "id_bits": id_bits}
        self.read_address = AxiAddressStream(addr_bits, data_bytes=self.data_bytes, **lite_args) @ DOWNWARDS
        self.read_data = AxiDataStream(data_bits, read=True, **lite_args) @ UPWARDS

        self.write_address = AxiAddressStream(addr_bits, data_bytes=self.data_bytes, **lite_args) @ DOWNWARDS
        self.write_data = AxiDataStream(data_bits, read=False, **lite_args) @ DOWNWARDS
        self.write_response = AxiWriteResponseStream(**lite_args) @ UPWARDS



from nmigen import *
from . import AxiEndpoint

__all__ = ["AxiFullToLiteBridge"]


class AxiFullToLiteBridge(Elaboratable):
    def __init__(self, full_master: AxiEndpoint):
        assert not full_master.is_lite
        self._full_master = full_master
        self.lite_master = AxiEndpoint.like(full_master, lite=True, name="axi_lite_bridge_master")

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self._full_master.connect_downstream(self.lite_master, allow_partial=True)
        
        # fake the id tracking
        read_id = Signal.like(self._full_master.read_data.id)
        write_id = Signal.like(self._full_master.write_data.id)

        with m.If(self._full_master.read_address.valid):
            m.d.comb += self._full_master.read_data.id.eq(self._full_master.read_address.id)
            m.d.sync += read_id.eq(self._full_master.read_address.id)
        with m.Else():
            m.d.comb += self._full_master.read_data.id.eq(read_id)

        with m.If(self._full_master.write_address.valid):
            m.d.comb += self._full_master.write_response.id.eq(self._full_master.write_address.id)
            m.d.sync += write_id.eq(self._full_master.write_address.id)
        with m.Else():
            m.d.comb += self._full_master.write_response.id.eq(write_id)

        m.d.comb += self._full_master.read_data.last.eq(1)

        return m



from .csr_types import *

from .fatbitstream import *
from .memorymap import *
from .peripheral import *
from .peripherals_aggregator import *
from .pydriver import *
from .devicetree_overlay import *
from .soc_platform import *
from .platform import *
from .cli import *



from nmigen import Elaboratable, Module, Signal, Array

from nap.cores import AxiReader, AxiWriter, if_none_get_zynq_hp_port, StreamInfo, LastWrapper, StreamTee
from nap import PacketizedStream, BasicStream, stream_transformer, StatusSignal

__all__ = ["DramPacketRingbufferStreamWriter", "DramPacketRingbufferStreamReader"]


class DramPacketRingbufferStreamWriter(Elaboratable):
    def __init__(
            self,
            input: PacketizedStream,
            max_packet_size, n_buffers, base_address=0x0f80_0000,
            axi=None,
    ):
        self.max_packet_size = max_packet_size
        self.base_address = base_address
        self.n_buffers = n_buffers
        self.axi = axi

        self.buffer_base_list = Array([base_address + max_packet_size * i for i in range(n_buffers)])
        self.buffer_level_list = Array([Signal(range(max_packet_size), name=f'buffer{i}_level') for i in range(n_buffers)])
        self.current_write_buffer = Signal(range(n_buffers))

        assert hasattr(input, "last")
        self.input = input

        self.overflowed_buffers = StatusSignal(32)
        self.buffers_written = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        axi = if_none_get_zynq_hp_port(self.axi, m, platform)
        assert len(self.input.payload) <= axi.data_bits

        tee = m.submodules.tee = StreamTee(self.input)

        data_stream = BasicStream(self.input.payload.shape())
        m.d.comb += data_stream.connect_upstream(tee.get_output(), allow_partial=True)

        transformer_input = tee.get_output()
        address_stream = BasicStream(axi.write_address.payload.shape())
        address_offset = Signal.like(axi.write_address.payload)
        is_in_overflow = Signal()
        stream_transformer(transformer_input, address_stream, m, latency=0, handle_out_of_band=False)
        with m.If(transformer_input.ready & transformer_input.valid):
            m.d.sync += self.buffer_level_list[self.current_write_buffer].eq(address_offset + axi.data_bytes)
            with m.If(transformer_input.last):
                m.d.sync += is_in_overflow.eq(0)
                next_buffer = (self.current_write_buffer + 1) % self.n_buffers
                m.d.sync += address_offset.eq(0)
                m.d.sync += self.current_write_buffer.eq(next_buffer)
                m.d.sync += self.buffers_written.eq(self.buffers_written + 1)
            with m.Else():
                with m.If((address_offset + axi.data_bytes < self.max_packet_size)):
                    m.d.sync += address_offset.eq(address_offset + axi.data_bytes)
                with m.Else():
                    with m.If(~is_in_overflow):
                        m.d.sync += is_in_overflow.eq(1)
                        m.d.sync += self.overflowed_buffers.eq(self.overflowed_buffers + 1)
        m.d.comb += address_stream.payload.eq(address_offset + self.buffer_base_list[self.current_write_buffer])

        m.submodules.writer = AxiWriter(address_stream, data_stream, axi)

        m.submodules.input_stream_info = StreamInfo(self.input)

        return m


class DramPacketRingbufferStreamReader(Elaboratable):
    def __init__(self, writer: DramPacketRingbufferStreamWriter, data_width=64, length_fifo_depth=1, axi=None):
        self.writer = writer
        self.data_width = data_width
        self.length_fifo_depth = length_fifo_depth
        self.axi = axi

        self.current_read_buffer = StatusSignal(self.writer.current_write_buffer.shape())

        self.output = PacketizedStream(data_width)

    def elaborate(self, platform):
        m = Module()

        axi = if_none_get_zynq_hp_port(self.axi, m, platform)
        assert len(self.output.payload) == axi.data_bits
        writer = self.writer

        address_stream = PacketizedStream(axi.read_address.payload.shape())
        m.submodules.address_stream_info = StreamInfo(address_stream)

        address_offset = Signal.like(axi.read_address.payload)

        with m.If(address_offset < writer.buffer_level_list[self.current_read_buffer]):
            m.d.comb += address_stream.valid.eq(1)
            m.d.comb += address_stream.payload.eq(address_offset + writer.buffer_base_list[self.current_read_buffer])
            m.d.comb += address_stream.last.eq(address_offset + axi.data_bytes >= writer.buffer_level_list[self.current_read_buffer])
            with m.If(address_stream.ready):
                m.d.sync += address_offset.eq(address_offset + axi.data_bytes)
        with m.Else():
            next_buffer = Signal.like(self.current_read_buffer)
            with m.If(writer.current_write_buffer == 0):
                m.d.comb += next_buffer.eq(writer.n_buffers - 1)
            with m.Else():
                m.d.comb += next_buffer.eq(writer.current_write_buffer - 1)
            m.d.sync += self.current_read_buffer.eq(next_buffer)
            m.d.sync += address_offset.eq(0)

        reader = m.submodules.axi_reader = LastWrapper(address_stream, lambda i: AxiReader(i, axi=axi, axi_data_width=len(self.output.payload)), last_fifo_depth=10, last_rle_bits=32)
        m.d.comb += self.output.connect_upstream(reader.output)

        m.submodules.output_stream_info = StreamInfo(self.output)

        return m



from nmigen import *
from nap.vendor import JTAG

__all__ = ["JTAGPeripheralConnector"]


class JTAGPeripheralConnector(Elaboratable):
    def __init__(self, peripheral, jtag=None, jtag_domain="jtag"):
        """
        A simple `PeripheralConnector` implementation for querying `Peripheral`s via JTAG in debug situations.
        This code does not handle memorymap stuff. Use in combination with PeripheralsAggregator
        """

        assert callable(peripheral.handle_read) and callable(peripheral.handle_write)
        assert isinstance(peripheral.range(), range)
        self.peripheral = peripheral
        self.jtag = jtag
        self.jtag_domain = jtag_domain

    def elaborate(self, platform):
        m = Module()
        if not self.jtag:
            jtag = m.submodules.jtag = JTAG(self.jtag_domain)
        else:
            jtag = self.jtag

        address_range = self.peripheral.range()
        state = Signal(range(10))

        addr = Signal(32)
        data = Signal(32)
        status = Signal()

        read_write_done = Signal()

        def read_write_done_callback(error):
            m.d.sync += status.eq(error)
            m.d.sync += read_write_done.eq(1)

        # This FSM should has two important properties:
        # 1) It can be controlled with an unknown latency on both tdi and tdo
        # 2) When it is in an unknown state we can return to a defined state (IDLE0) by shifting in "0" for 96 cycles
        # READ:
        # <0>*<1>(ADDRESS[32])<0><1 keep high>
        #                                             <1>(DATA[32])(STATUS_BIT[1])
        # WRITE:
        # <0>*<1>(ADDRESS[32])<1>(DATA[32])
        #                                             <1>(STATUS_BIT[1])
        with m.FSM():
            def next_on_jtag_shift(next_state, use_tdo_shift=False):
                with m.If(jtag.shift_tdo if use_tdo_shift else jtag.shift_tdi):
                    m.next = next_state

            with m.State("IDLE0"):
                m.d.comb += state.eq(0)
                m.d.sync += read_write_done.eq(0)
                with m.If(~jtag.tdi):  # in IDLE0 IDLE1 we wait for a 0->1 transition to have property 2) (see above)
                    next_on_jtag_shift("IDLE1")
            with m.State("IDLE1"):
                m.d.comb += state.eq(1)
                with m.If(jtag.tdi):
                    next_on_jtag_shift("ADDR0")

            # address states
            for i in range(32):
                with m.State("ADDR{}".format(i)):
                    m.d.comb += state.eq(2)
                    m.d.sync += addr[i].eq(jtag.tdi)
                    if i < 31:
                        next_on_jtag_shift("ADDR{}".format(i + 1))
                    else:
                        next_on_jtag_shift("RW_CMD")

            with m.State("RW_CMD"):  # we recive one bit that indicates if we want to read (0) or write (1)
                m.d.comb += state.eq(3)
                with m.If(jtag.tdi):
                    next_on_jtag_shift("WRITE0")
                with m.Else():
                    next_on_jtag_shift("READ_WAIT")

            # read states
            with m.State("READ_WAIT"):
                m.d.comb += state.eq(4)
                self.peripheral.handle_read(m, addr - address_range.start, data, read_write_done_callback)
                with m.If(read_write_done):
                    m.d.comb += jtag.tdo.eq(1)
                    next_on_jtag_shift("READ0")
                with m.If(~jtag.tdi):  # we are requested to abort the waiting
                    next_on_jtag_shift("IDLE0")
            for i in range(32):
                with m.State("READ{}".format(i)):
                    m.d.comb += state.eq(5)
                    m.d.comb += jtag.tdo.eq(data[i])
                    if i < 31:
                        next_on_jtag_shift("READ{}".format(i + 1), use_tdo_shift=True)
                    else:
                        next_on_jtag_shift("READ_STATUS")
            with m.State("READ_STATUS"):
                m.d.comb += state.eq(6)
                m.d.comb += jtag.tdo.eq(status)
                next_on_jtag_shift("IDLE0")

            # write states
            for i in range(32):
                with m.State("WRITE{}".format(i)):
                    m.d.comb += state.eq(7)
                    m.d.sync += data[i].eq(jtag.tdi)
                    if i < 31:
                        next_on_jtag_shift("WRITE{}".format(i + 1))
                    else:
                        next_on_jtag_shift("WRITE_WAIT")
            with m.State("WRITE_WAIT"):
                m.d.comb += state.eq(8)
                self.peripheral.handle_write(m, addr - address_range.start, data, read_write_done_callback)
                with m.If(read_write_done):
                    m.d.comb += jtag.tdo.eq(1)
                    next_on_jtag_shift("WRITE_STATUS")
                with m.If(~jtag.tdi):  # we are requested to abort the waiting
                    next_on_jtag_shift("IDLE0")
            with m.State("WRITE_STATUS"):
                m.d.comb += state.eq(9)
                m.d.comb += jtag.tdo.eq(status)
                next_on_jtag_shift("IDLE0")

        if self.jtag_domain != "sync":
            return DomainRenamer(self.jtag_domain)(m)
        else:
            return m



# a helper to generate the plugin connector *on the camera side*
from nmigen.build import Connector, DiffPairs
from nap.util.sim import SimPlatform

__all__ = ["add_plugin_connector", "PluginDiffPair", "is_resource_pin_inverted", "is_connector_pin_inverted", "is_signal_inverted"]


def add_plugin_connector(platform, number, conn=None, lvds=None, gpio=None, i2c=None):
    if not hasattr(platform, "inverts"):
        platform.inverts = {}

    mapping = {}

    if lvds is None:
        lvds = []
    for i, l in enumerate(lvds):
        splited = l.split()
        if len(splited) == 3:
            p, n = splited[0], splited[2]
            invert = True
        elif len(splited) == 2:
            p, n = splited
            invert = False
        else:
            raise ValueError()

        platform.inverts["plugin_{}:lvds{}_p".format(number, i)] = invert
        mapping["lvds{}_p".format(i)] = "{}".format(p)
        mapping["lvds{}_n".format(i)] = "{}".format(n)

    if gpio is None:
        gpio = []
    for i, g in enumerate(gpio):
        mapping["gpio{}".format(i)] = "{}".format(g)

    if i2c is not None:
        scl, sda = i2c

        mapping["i2c_scl"] = "{}".format(scl)
        mapping["i2c_sda"] = "{}".format(sda)

    platform.add_connectors([
        Connector(
            "plugin", number,
            mapping,
            conn=conn,
        ),
    ])


# the serdes parameter makes that we do not invert the pin here but defer it
def _is_pin_inverted(platform, plugin_number, pin, serdes):
    if not hasattr(platform, "inverts"):
        return False

    inverts_intex = "plugin_{}:lvds{}_p".format(plugin_number, pin)
    invert = platform.inverts[inverts_intex]
    if serdes:
        return False
    else:
        del platform.inverts[inverts_intex]
        return invert


def PluginDiffPair(platform, plugin_number, pin, dir, serdes=False):
    return DiffPairs(
        p="lvds{}_p".format(pin),
        n="lvds{}_n".format(pin),
        dir=dir,
        conn=("plugin", plugin_number),
        invert=_is_pin_inverted(platform, plugin_number, pin, serdes)
    )


def is_connector_pin_inverted(platform, connector_pin):
    if not hasattr(platform, "inverts"):
        return False

    if connector_pin in platform.inverts:
        invert = platform.inverts[connector_pin]
        return invert
    else:
        return False


def is_resource_pin_inverted(platform, resource, path):
    def subsignal_by_path(subsignal, path):
        for s in subsignal.ios:
            if s.name == path[0]:
                if len(path) > 1:
                    return subsignal_by_path(s, path[1:])
                else:
                    return s
        raise KeyError()
    subsignal = subsignal_by_path(platform.resources[resource], path)
    assert len(subsignal.ios) == 1
    assert len(subsignal.ios[0].p.names) == 1
    return is_connector_pin_inverted(platform, subsignal.ios[0].p.names[0])


def is_signal_inverted(platform, signal):
    if isinstance(platform, SimPlatform):
        return False

    for res_name, res in platform._requested.items():
        def recurse_fields(x, path):
            if hasattr(x, "fields"):
                for pin, field in x.fields.items():
                    if field is signal:
                        return is_resource_pin_inverted(platform, res_name, [*path, pin])
                    result = recurse_fields(field, path=[*path, pin])
                    if result is not None:
                        return result
            return None
        result = recurse_fields(res, path=[])
        if result is not None:
            break
    if result is not None:
        return result
    raise KeyError("signal {} was not found in platform. could not determine if it should be inverted".format(signal))



from itertools import product
from nmigen import *
from nmigen.build import Clock
from nap import ControlSignal, StatusSignal, max_error_freq
from nap.cores import RGB24
from nap.vendor.xilinx_s7 import Mmcm, OSerdes10, PS7
from .parse_modeline import parse_modeline
from .tmds import TmdsEncoder

__all__ = ["HdmiTx", "HdmiClocking", "HdmiTimingGenerator", "HdmiPluginLowspeedController"]


class HdmiTx(Elaboratable):
    def __init__(self, resource, modeline, pix_domain="pix", generate_clocks=True):
        self.resource = resource
        self.pix_domain = pix_domain
        self.generate_clocks = generate_clocks
        self.initial_video_timing = parse_modeline(modeline)
        self.pix_freq = Clock(self.initial_video_timing.pxclk * 1e6)

        self.rgb = RGB24()

        self.hsync_polarity = ControlSignal()
        self.vsync_polarity = ControlSignal()

        self.clock_pattern = ControlSignal(10, name="hdmi_clock_pattern", reset=0b1111100000)

        self.timing_generator = HdmiTimingGenerator(self.initial_video_timing)

    def elaborate(self, platform):
        m = Module()
        if self.generate_clocks:
            self.clocking = m.submodules.clocking = HdmiClocking(self.pix_freq, self.pix_domain)

        in_pix_domain = DomainRenamer(self.pix_domain)
        timing = m.submodules.timing_generator = in_pix_domain(self.timing_generator)

        domain_args = {"domain": self.pix_domain, "domain_5x": "{}_5x".format(self.pix_domain)}

        control_char = Signal(2)
        m.d.comb += control_char[0].eq(timing.hsync ^ self.hsync_polarity)
        m.d.comb += control_char[1].eq(timing.vsync ^ self.vsync_polarity)

        encoder_r = m.submodules.encoder_r = in_pix_domain(TmdsEncoder(self.rgb.r, control_char, timing.active))
        encoder_g = m.submodules.encoder_g = in_pix_domain(TmdsEncoder(self.rgb.g, control_char, timing.active))
        encoder_b = m.submodules.encoder_b = in_pix_domain(TmdsEncoder(self.rgb.b, control_char, timing.active))

        serializer_clock = m.submodules.serializer_clock = OSerdes10(self.clock_pattern, self.resource.clock, **domain_args)
        serializer_b = m.submodules.serializer_b = OSerdes10(encoder_b.out, self.resource.b, **domain_args)
        serializer_g = m.submodules.serializer_g = OSerdes10(encoder_g.out, self.resource.g, **domain_args)
        serializer_r = m.submodules.serializer_r = OSerdes10(encoder_r.out, self.resource.r, **domain_args)

        m.submodules.lowspeed = HdmiPluginLowspeedController(self.resource)

        return m


class HdmiClocking(Elaboratable):
    def __init__(self, pix_freq, pix_domain):
        self.pix_domain = pix_domain
        self.pix_freq = pix_freq

    def find_valid_config(self):
        # fast path:
        for output_div in [2, 1, 4, 6]:
            for fclk_frequency in [10e6, 20e6]:
                for mmcm_mul in sorted(Mmcm.vco_multipliers, key=lambda a: abs(a - 39)):
                    if (fclk_frequency * mmcm_mul == self.pix_freq.frequency * 5 * output_div) \
                            and Mmcm.is_valid_vco_conf(fclk_frequency, mmcm_mul, 1):
                        self.mmcm_mul = mmcm_mul
                        self.fclk_freq = fclk_frequency
                        self.output_div = output_div
                        self.mmcm_div = 1
                        return

        # always working path:
        print("WARNING: falling back to slow MMCM calculation path")
        valid_configs = (
            (abs(1 - ((fclk * mul / div) / (self.pix_freq.frequency * 5 * output_div))),
             fclk, mul, div, output_div)
            for fclk, mul, div, output_div in product(
            [f for f in PS7.get_possible_fclk_frequencies() if 1e6 <= f <= 100e6],
            Mmcm.vco_multipliers,
            [1],
            range(1, 6)
        )
            if Mmcm.is_valid_vco_conf(fclk, mul, div)
        )
        deviation, self.fclk_freq, self.mmcm_mul, self.mmcm_div, self.output_div = sorted(valid_configs)[0]

    def elaborate(self, platform):
        m = Module()

        self.find_valid_config()

        platform.ps7.fck_domain(self.fclk_freq, "{}_synth_fclk".format(self.pix_domain))
        mmcm = m.submodules.mmcm = Mmcm(
            input_clock=self.fclk_freq, input_domain="{}_synth_fclk".format(self.pix_domain),
            vco_mul=self.mmcm_mul, vco_div=self.mmcm_div
        )

        deviation = (abs(
            1 - ((self.fclk_freq * self.mmcm_mul / self.mmcm_div) / (self.pix_freq.frequency * 5 * self.output_div))))
        print("pixclk {}Mhz, error: {}%. (fclk={}MHz; mul={}; div={}; output_div={})".format(
            self.pix_freq.frequency / 1e6, deviation, self.fclk_freq / 1e6, self.mmcm_mul, self.mmcm_div,
            self.output_div
        ))

        actual_pix_freq = mmcm.output_domain(self.pix_domain, 5 * self.output_div)
        max_error_freq(actual_pix_freq.frequency, self.pix_freq.frequency)

        mmcm.output_domain("{}_5x".format(self.pix_domain), self.output_div)

        return m


class HdmiTimingGenerator(Elaboratable):
    def __init__(self, video_timing, vertical_signals_shape=range(8000), horizontal_signals_shape=range(4000)):
        self.hscan = ControlSignal(horizontal_signals_shape, reset=video_timing.hscan)
        self.vscan = ControlSignal(vertical_signals_shape, reset=video_timing.vscan)
        self.width = ControlSignal(horizontal_signals_shape, reset=video_timing.hres)
        self.height = ControlSignal(vertical_signals_shape, reset=video_timing.vres)
        self.hsync_start = ControlSignal(horizontal_signals_shape, reset=video_timing.hsync_start)
        self.hsync_end = ControlSignal(horizontal_signals_shape, reset=video_timing.hsync_end)
        self.vsync_start = ControlSignal(vertical_signals_shape, reset=video_timing.vsync_start)
        self.vsync_end = ControlSignal(vertical_signals_shape, reset=video_timing.vsync_end)

        self.x = StatusSignal(horizontal_signals_shape,)
        self.y = StatusSignal(vertical_signals_shape)
        self.active = StatusSignal()
        self.is_blanking_x = StatusSignal()
        self.is_blanking_y = StatusSignal()
        self.hsync = StatusSignal()
        self.vsync = StatusSignal()

    def elaborate(self, plat):
        m = Module()

        # set the xy coordinates
        with m.If(self.x < self.hscan - 1):
            m.d.sync += self.x.eq(self.x + 1)
        with m.Else():
            m.d.sync += self.x.eq(0)
            with m.If(self.y < self.vscan - 1):
                m.d.sync += self.y.eq(self.y + 1)
            with m.Else():
                m.d.sync += self.y.eq(0)

        m.d.comb += [
            self.is_blanking_x.eq(self.x >= self.width),
            self.is_blanking_y.eq(self.y >= self.height),
            self.active.eq((self.x < self.width) & (self.y < self.height)),
            self.hsync.eq((self.x >= self.hsync_start) & (self.x < self.hsync_end)),
            self.vsync.eq((self.y >= self.vsync_start) & (self.y < self.vsync_end))
        ]

        return m


class HdmiPluginLowspeedController(Elaboratable):
    def __init__(self, plugin):
        self.plugin = plugin

    def elaborate(self, platform):
        m = Module()

        if hasattr(self.plugin, "output_enable"):  # micro style directly connected hdmi plugin module
            sigs = [
                ("output_enable", ControlSignal, 1, 1),
                ("equalizer", ControlSignal, 2, 0b11),
                ("dcc_enable", ControlSignal, 1, 0),
                ("vcc_enable", ControlSignal, 1, 1),
                ("ddet", ControlSignal, 1, 0),
                ("ihp", StatusSignal, 1, 0),
            ]
        elif hasattr(self.plugin, "out_en"):  # zybo style raw hdmi
            sigs = ("out_en", ControlSignal, 1, 1),
        else:
            sigs = []

        # generate low speed CSR signals
        for name, signal_type, width, default in sigs:
            csr_signal = signal_type(width, reset=default)
            setattr(self, name, csr_signal)
            io = getattr(self.plugin, name)
            if signal_type == ControlSignal:
                m.d.comb += io.eq(csr_signal)
            else:
                m.d.comb += csr_signal.eq(io)

        return m



import shlex
from dataclasses import dataclass

__all__ = ['VideoTiming', 'parse_modeline']


@dataclass
class VideoTiming:
    pxclk: float
    hres: int
    hsync_start: int
    hsync_end: int
    hscan: int
    vres: int
    vsync_start: int
    vsync_end: int
    vscan: int


def parse_modeline(modeline: str):
    assert modeline.startswith("Modeline")
    modeline = shlex.split(modeline)

    pxclk = float(modeline[2])

    names = ["hres", "hsync_start", "hsync_end", "hscan", "vres", "vsync_start", "vsync_end", "vscan"]
    values = map(int, modeline[3:-2])

    return_dict = dict(zip(names, values))
    return_dict["pxclk"] = pxclk

    return VideoTiming(**return_dict)



from nmigen import *
from nap import StatusSignal, ControlSignal
from nap.cores import StreamCombiner, StreamInfo, InflexibleSourceDebug, ImageStream
from nap.util import delay_by, ends_with
from .s7_phy import HispiPhy

__all__ = ["HispiRx"]


# those are only the starts of the patterns; they are expanded to the length of the word
control_words = {
    "START_OF_ACTIVE_FRAME_IMAGE_DATA": "00011",
    "START_OF_ACTIVE_FRAME_EMBEDDED_DATA": "10011",
    "START_OF_ACTIVE_LINE_IMAGE_DATA": "00001",
    "START_OF_ACTIVE_LINE_EMBEDDED_DATA": "10001",
    "END_OF_ACTIVE_FRAME": "111",
    "END_OF_ACTIVE_LINE": "101",
    "START_OF_VERTICAL_BLANKING_LINE": "1001",
}


class LaneManager(Elaboratable):
    def __init__(self, input_data: Signal, sync_pattern=(-1, 0, 0)):
        """
        Aligns the word boundries of one Hispi lane and detects control codes.
        Compatible only with Packetized-SP mode because it needs end markers.

        :param sync_pattern: the preamble of a control word (default is correct for most cases)
        :param timeout: Issue a bit slip after a control word wasnt found for n cycles
        """
        self.sync_pattern = sync_pattern
        self.input_data = input_data

        self.is_aligned = StatusSignal()

        self.timeout = ControlSignal(32, reset=10000)
        self.timeouts_to_resync = ControlSignal(32, reset=10000)

        self.since_last_sync_pattern_or_bitslip = StatusSignal(32)
        self.performed_bitslips = StatusSignal(32)
        self.timeouts_since_alignment = StatusSignal(32)
        self.last_word = StatusSignal(input_data.shape())
        self.last_control_word = StatusSignal(
            input_data.shape(),
            decoder=lambda x: next((
                "{}/{:012b}".format(control_word, x)
                for control_word, ending
                in control_words.items()
                if "{:012b}".format(x).endswith(ending)
            ), "UNKNOWN/{:012b}".format(x))
        )

        self.do_bitslip = Signal()
        self.output = ImageStream(self.input_data.shape())

    def elaborate(self, platform):
        m = Module()

        m.d.sync += self.last_word.eq(self.input_data)
        m.d.comb += self.output.payload.eq(self.input_data)

        with m.If(self.since_last_sync_pattern_or_bitslip < self.timeout):
            m.d.sync += self.since_last_sync_pattern_or_bitslip.eq(self.since_last_sync_pattern_or_bitslip + 1)
        with m.Else():
            with m.If(self.is_aligned):
                with m.If(self.timeouts_since_alignment > self.timeouts_to_resync):
                    m.d.sync += self.is_aligned.eq(0)
                    m.d.sync += self.timeouts_since_alignment.eq(0)
                with m.Else():
                    m.d.sync += self.timeouts_since_alignment.eq(self.timeouts_since_alignment + 1)
            with m.Else():
                m.d.sync += self.since_last_sync_pattern_or_bitslip.eq(0)
                m.d.sync += self.performed_bitslips.eq(self.performed_bitslips + 1)
                m.d.comb += self.do_bitslip.eq(1)

        with m.FSM():
            for i, pattern_word in enumerate(self.sync_pattern):
                with m.State("sync{}".format(i)):
                    with m.If(self.input_data == Const(pattern_word, self.input_data.shape())):
                        if i < len(self.sync_pattern) - 1:
                            m.next = "sync{}".format(i + 1)
                        else:
                            m.next = "control_word"
                    with m.Else():
                        with m.If(self.input_data == Const(self.sync_pattern[0], self.input_data.shape())):
                            m.next = "sync1"
                        with m.Else():
                            m.next = "sync0"
            with m.State("control_word"):
                with m.If(ends_with(self.input_data, *control_words.values())):
                    m.d.sync += self.last_control_word.eq(self.input_data)
                    m.d.sync += self.since_last_sync_pattern_or_bitslip.eq(0)
                    m.d.sync += self.is_aligned.eq(1)
                m.next = "sync0"

        # assemble the output stream
        valid = Signal()
        m.d.comb += valid.eq(ends_with(
            self.last_control_word,
            # control_words["START_OF_ACTIVE_FRAME_EMBEDDED_DATA"],
            control_words["START_OF_ACTIVE_FRAME_IMAGE_DATA"],
            # control_words["START_OF_ACTIVE_LINE_EMBEDDED_DATA"],
            control_words["START_OF_ACTIVE_LINE_IMAGE_DATA"]
        ))

        # delay is needed because we only know that the line finished when the control code is done
        # this is len(sync_pattern) + 1 + 1 cycles after the line really ended
        delayed_valid = delay_by(valid, len(self.sync_pattern) + 2, m)
        delayed_data = delay_by(self.input_data, len(self.sync_pattern) + 2, m)

        with m.If(ends_with(self.last_control_word,
                            control_words["END_OF_ACTIVE_FRAME"], control_words["END_OF_ACTIVE_LINE"])):
            m.d.sync += delayed_valid.eq(0)

        m.d.comb += self.output.payload.eq(delayed_data)
        m.d.comb += self.output.valid.eq(delayed_valid)
        m.d.comb += self.output.frame_last.eq(ends_with(self.last_control_word, control_words["END_OF_ACTIVE_FRAME"]))
        m.d.comb += self.output.line_last.eq(ends_with(self.last_control_word, control_words["END_OF_ACTIVE_LINE"], control_words["END_OF_ACTIVE_FRAME"]))

        m.submodules.debug = InflexibleSourceDebug(self.output)

        return m


class HispiRx(Elaboratable):
    def __init__(self, sensor, bits=12, hispi_domain="hispi"):
        self.hispi_domain = hispi_domain
        self.lvds_clk = sensor.lvds_clk
        self.lvds = sensor.lvds
        self.lanes = len(self.lvds)
        self.bits = bits

        self.output = ImageStream(len(self.lvds) * bits)
        self.output_domain = hispi_domain
        self.phy = HispiPhy(num_lanes=self.lanes, bits=self.bits, hispi_domain=hispi_domain)

    def elaborate(self, platform):
        m = Module()

        phy = m.submodules.phy = self.phy
        m.d.comb += phy.hispi_clk.eq(self.lvds_clk)
        m.d.comb += phy.hispi_lanes.eq(self.lvds)

        in_hispi_domain = DomainRenamer(self.hispi_domain)

        streams = []
        for i, lane in enumerate(phy.out):
            lane_manager = m.submodules["lane_manager_{}".format(i)] = in_hispi_domain(LaneManager(lane))

            m.d.comb += phy.bitslip[i].eq(lane_manager.do_bitslip)

            m.d.comb += self.output.payload[i * self.bits: (i + 1) * self.bits].eq(lane_manager.input_data)
            streams.append(lane_manager.output)

        stream_combiner = m.submodules.stream_combiner = in_hispi_domain(StreamCombiner(*streams, merge_payload=True))
        m.submodules.output_stream_info = in_hispi_domain(StreamInfo(stream_combiner.output))
        m.d.comb += self.output.connect_upstream(stream_combiner.output)

        return m



# An experiment that allows debugging / diagnosing performance of the FT601 USB3 FIFO ic
from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        pass

    def elaborate(self, platform):
        m = Module()

        ft601 = platform.request("ft601")
        ft601_perf_debug = m.submodules.ft601_perf_debug = FT601PerfDebug(ft601)
        m.d.comb += platform.request("led", 0).eq(1)

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(Usb3PluginPlatform, HdmiDigitizerPlatform), possible_socs=(JTAGSocPlatform, ZynqSocPlatform))



from nmigen import *

from nap import ControlSignal, stream_transformer, nAvrg
from . import ImageStream, RGB24, ImageConvoluter

__all__ = ["RecoloringDebayerer", "SimpleInterpolatingDebayerer"]


class RecoloringDebayerer(Elaboratable):
    """Debayer an image by simply coloring the pixels in the correct color"""
    def __init__(self, input: ImageStream):
        self.input = input
        self.output = ImageStream(24)

        self.shift_x = ControlSignal()
        self.shift_y = ControlSignal()

    def elaborate(self, platform):
        m = Module()

        x_odd = Signal()
        y_odd = Signal()
        stream_transformer(self.input, self.output, m, latency=0)
        with m.If(self.input.ready & self.input.valid):
            with m.If(self.input.line_last):
                m.d.sync += x_odd.eq(0)
                m.d.sync += y_odd.eq(~y_odd)
            with m.Else():
                m.d.sync += x_odd.eq(~x_odd)
            with m.If(self.input.frame_last):
                m.d.sync += y_odd.eq(0)
                m.d.sync += x_odd.eq(0)

        x = x_odd ^ self.shift_x
        y = y_odd ^ self.shift_y

        rgb = RGB24()
        with m.If(x & ~y):
            m.d.comb += rgb.r.eq(self.input.payload)
        with m.Elif(~x & y):
            m.d.comb += rgb.b.eq(self.input.payload)
        with m.Else():
            m.d.comb += rgb.g.eq(self.input.payload // 2)

        m.d.comb += self.output.payload.eq(rgb)

        return m


class SimpleInterpolatingDebayerer(Elaboratable):
    """Debayer an image by interpolating the colour with the neighbouring pixels"""
    def __init__(self, input: ImageStream, width=3000, height=3000):
        self.input = input
        self.output = ImageStream(24)

        self.width = width
        self.height = height

        self.shift_x = ControlSignal()
        self.shift_y = ControlSignal()

    def elaborate(self, platform):
        m = Module()

        def transformer_function(x, y, image_proxy):
            x_even = Signal()
            m.d.comb += x_even.eq((x + self.shift_x) % 2 == 0)
            y_even = Signal()
            m.d.comb += y_even.eq((y + self.shift_y) % 2 == 0)

            rgb = RGB24()
            with m.If(x_even & ~y_even):  # we are a red pixel
                m.d.comb += rgb.r.eq(image_proxy[x, y])
                m.d.comb += rgb.g.eq(nAvrg(
                    (image_proxy[x-1, y]),
                    (image_proxy[x+1, y]),
                    (image_proxy[x, y-1]),
                    (image_proxy[x, y+1]),
                ))
                m.d.comb += rgb.b.eq(nAvrg(
                    (image_proxy[x - 1, y - 1]),
                    (image_proxy[x + 1, y + 1]),
                    (image_proxy[x + 1, y - 1]),
                    (image_proxy[x - 1, y + 1]),
                ))
            with m.Elif(~x_even & y_even):  # we are a blue pixel
                m.d.comb += rgb.b.eq(image_proxy[x, y])
                m.d.comb += rgb.g.eq(nAvrg(
                    (image_proxy[x - 1, y]),
                    (image_proxy[x + 1, y]),
                    (image_proxy[x, y - 1]),
                    (image_proxy[x, y + 1]),
                ))
                m.d.comb += rgb.r.eq(nAvrg(
                    (image_proxy[x - 1, y - 1]),
                    (image_proxy[x + 1, y + 1]),
                    (image_proxy[x + 1, y - 1]),
                    (image_proxy[x - 1, y + 1]),
                ))
            with m.Elif(~x_even & ~y_even):  # we are a green pixel in a red row
                m.d.comb += rgb.g.eq(image_proxy[x, y])
                m.d.comb += rgb.r.eq(nAvrg(
                    (image_proxy[x - 1, y]),
                    (image_proxy[x + 1, y]),
                ))
                m.d.comb += rgb.b.eq(nAvrg(
                    (image_proxy[x, y - 1]),
                    (image_proxy[x, y + 1]),
                ))
            with m.Elif(x_even & y_even):  # we are a green pixel in a blue row
                m.d.comb += rgb.g.eq(image_proxy[x, y])
                m.d.comb += rgb.b.eq(nAvrg(
                    (image_proxy[x - 1, y]),
                    (image_proxy[x + 1, y]),
                ))
                m.d.comb += rgb.r.eq(nAvrg(
                    (image_proxy[x, y - 1]),
                    (image_proxy[x, y + 1]),
                ))

            return rgb

        video_transformer = m.submodules.video_transformer = ImageConvoluter(self.input, transformer_function,
                                                                              self.width, self.height)
        m.d.comb += self.output.connect_upstream(video_transformer.output)

        return m


from nmigen import *
from nap import BasicStream, ControlSignal
from nap.vendor.xilinx_s7 import DDRSerializer

__all__ = ["MipiMultiLaneTxPhy"]


class MipiMultiLaneTxPhy(Elaboratable):
    def __init__(self, input: BasicStream, clock_pin, lane_pins, ddr_domain):
        self.input = input
        self.clock_pin = clock_pin
        self.lane_pins = lane_pins
        self.ddr_domain = ddr_domain
        self.clock_pattern = ControlSignal(8, reset=0b00001111)

        assert len(input.payload) == len(lane_pins) * 8

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.input.ready.eq(1)
        m.submodules.clock_tx = DDRSerializer(self.clock_pattern, self.clock_pin, self.ddr_domain, bit_width=8, msb_first=False)
        for i, lane_pin in enumerate(self.lane_pins):
            m.submodules[f'lane{i}_phy'] = DDRSerializer(self.input.payload[i * 8: (i + 1) * 8], lane_pin, self.ddr_domain, bit_width=8, msb_first=False)

        return m



# Utils for generating "fatbitstreams" (files that contain loading-logic, a bitstream, initialization and maybe drivers)
from base64 import b64encode
from datetime import datetime
from shlex import quote

__all__ = ["FatbitstreamContext"]


class FatbitstreamContext:
    platform_to_context_dict = {}

    @classmethod
    def get(cls, platform):
        """
        Implements a singleton-like pattern here to have exactly one FatbitstreamContext per Platform instance
        :rtype: FatbitstreamContext
        """
        if hasattr(platform, "_wrapped_platform"):  # we cant check directly for isinstance SocPlatform here because of cyclic imports
            platform = platform._wrapped_platform

        if platform not in cls.platform_to_context_dict:
            cls.platform_to_context_dict[platform] = FatbitstreamContext(platform, called_by_get_classmethod=True)
        return cls.platform_to_context_dict[platform]

    def __init__(self, platform, called_by_get_classmethod=False):
        assert called_by_get_classmethod
        self.self_extracting_blobs = {}  # a dict of filename -> contents
        self.init_commands = []
        self.pre_init_commands = []
        self._platform = platform

    def generate_fatbitstream_generator(self, build_name):
        """
        Generates shell code that is inserted into the nmigen build file that generates the bash file.
        BEWARE: this has kind of a double indirection because we write bash that generates bash that is then executed.
        Yields the commands that are to be executed during build time.
        """
        builder = _FatbitstreamBuilder()
        builder.append_host("\n")
        builder.append_host("touch {{name}}.fatbitstream.sh")
        builder.append_command("#!/bin/bash\n")
        builder.append_command('"# Fatbitstream for "{}" on platform {} with soc {}. Build on $(date)\n"'.format(
            build_name, self._platform.__class__.__name__, self._platform._soc_platform.__class__.__name__,
            datetime.now().strftime("%d.%b.%Y %H:%M:%S")
        ), do_quote=False)
        builder.append_command("set -euo pipefail\n")

        # we create a directory for all the files we will eventually unpack to allow multiple fatbitstreams
        # to coexist in the same top directory. this is for example important when we want to load a plugin module
        # (with JTAGSoc) in parallel with a Zynq bitstream.
        builder.append_command("mkdir -p '{{name}}.fatbitstream.d'\n")
        builder.append_command("cd '{{name}}.fatbitstream.d'\n")

        builder.append_command("\n\n### driver unpacking ###\n")
        for filename, contents in self.self_extracting_blobs.items():
            builder.append_self_extracting_blob_from_string(contents, filename)

        builder.append_command("\n\n### pre init script ###\n")
        for cmd in self.pre_init_commands:
            builder.append_command(cmd + "\n")

        assert hasattr(self._platform._soc_platform, "pack_bitstream_fatbitstream")
        builder.append_command("\n\n### fpga bitstream loading  ###\n")
        self._platform._soc_platform.pack_bitstream_fatbitstream(builder)

        builder.append_command("\n\n### init script ###\n")
        for cmd in self.init_commands:
            builder.append_command(cmd + "\n")

        builder.append_host("chmod +x {{name}}.fatbitstream.sh")

        return builder.cmds


class _FatbitstreamBuilder:
    def __init__(self):
        self.cmds = []

    def append_host(self, cmd):
        self.cmds.append(cmd)

    def append_command(self, cmd, do_quote=True):
        self.append_host("echo -ne {} >> {{{{name}}}}.fatbitstream.sh".format(quote(cmd.replace("\n", "\\n")) if do_quote else cmd.replace("\n", "\\n")))

    def append_self_extracting_blob_from_string(self, string, path):
        self.append_command("base64 -d > {} <<EOF\n{}\nEOF\n".format(quote(path), b64encode(string.encode("utf-8")).decode("ASCII")))

    def append_self_extracting_blob_from_file(self, src_path, target_path):
        self.append_command(
            '"base64 -d > {} <<EOF\\n$(base64 -w0 {})\\nEOF\\n"'.format(quote(target_path), quote(src_path)),
            do_quote=False
        )



from types import MethodType

from nmigen import *
from nmigen.compat import Module as CompatModule
from nmigen.hdl.xfrm import TransformedElaboratable


class ElaboratableSames:
    def __init__(self):
        self.sames = []

    def insert(self, a, b):
        for row in self.sames:
            if a in row and b not in row:
                row.append(b)
                return
            elif b in row and a not in row:
                row.append(a)
                return
            elif b in row and a in row:
                return
        self.sames.append([a, b])

    def get_row(self, something):
        candidates = list(row for row in self.sames if something in row)
        if len(candidates) == 0:
            raise AssertionError()
        if len(candidates) != 1:
            raise AssertionError()
        return candidates[0]

    def get_by_filter(self, something, item_filter):
        if isinstance(something, Instance):
            return None
        if self.get_row(something) is None:
            return None
        candidates = list(item for item in self.get_row(something) if item_filter(item))
        if len(candidates) == 0:
            return None
        if len(candidates) > 1:
            raise AssertionError()
        return candidates[0]

    def get_module(self, something):
        return self.get_by_filter(something, lambda x: isinstance(x, Module))

    def get_elaboratable(self, something):
        return self.get_by_filter(
            something, lambda x: isinstance(x, Elaboratable)
                                 and not isinstance(x, Module)
                                 and not isinstance(x, TransformedElaboratable)
        )


def inject_elaborate_wrapper(obj, sames):
    if hasattr(obj, 'elaborate') and not isinstance(obj, Fragment):

        def generate_elaborate_wrapper(real_elaborate):
            def elaborate_wrapper(self, platform):
                elaborated = real_elaborate(platform)
                # print("{} ({}) elaborated to {} ({})".format(
                #     self.__class__.__name__, self,
                #     elaborated.__class__.__name__, elaborated
                # ))
                sames.insert(self, elaborated)
                inject_elaborate_wrapper(elaborated, sames)
                return elaborated
            return elaborate_wrapper

        obj.elaborate = MethodType(generate_elaborate_wrapper(obj.elaborate), obj)
    else:
        if not isinstance(obj, (Instance, Fragment)):
            raise AssertionError()

    if isinstance(obj, Module):
        submodules = [*obj._named_submodules.values(), *obj._anon_submodules]
        for elab in submodules:
            inject_elaborate_wrapper(elab, sames)
    elif isinstance(obj, CompatModule):
        submodules = [submod for name, submod in obj.submodules._cm._submodules]
        for elab in submodules:
            inject_elaborate_wrapper(elab, sames)
    elif isinstance(obj, TransformedElaboratable):
        inject_elaborate_wrapper(obj._elaboratable_, sames)
        for i in range(len(obj._transforms_)):

            def generate_transform_wrapper(real_transform):
                def transform_wrapper(self, value, *, src_loc_at=0):
                    output = real_transform(self, value, src_loc_at=src_loc_at)
                    sames.insert(value, output)
                    # print("{} ({}) transformed to {} ({})".format(
                    #     value.__class__.__name__, value,
                    #     output.__class__.__name__, output
                    # ))
                    return output
                return transform_wrapper

            obj._transforms_[i].__class__.__call__ = generate_transform_wrapper(obj._transforms_[i].__class__.__call__)
        sames.insert(obj, obj._elaboratable_)
    else:
        if not (isinstance(obj, (Fragment, Instance)) or hasattr(obj, 'elaborate')):
            raise AssertionError()


def fragment_get_with_elaboratable_trace(elaboratable, platform, sames=None):
    # this is a hack to retrieve the elaboratable which produced a specific module
    if sames is None:
        sames = ElaboratableSames()

    inject_elaborate_wrapper(elaboratable, sames)
    fragment = Fragment.get(elaboratable, platform)

    return fragment, sames



from nmigen import *
from nap import BasicStream, ControlSignal

__all__ = ["CounterStreamSource"]


class CounterStreamSource(Elaboratable):
    def __init__(self, width, count_if_not_ready=False):
        self.output = BasicStream(width, name="counter_stream")

        self.count_if_not_ready = ControlSignal(reset=count_if_not_ready)

    def elaborate(self, platform):
        m = Module()

        # we initialize our counter with one to compensate for the 2 cycle delay
        # from it to the payload
        counter = Signal(self.output.payload.shape(), reset=1)

        m.d.comb += self.output.valid.eq(1)
        with m.If(self.output.ready | self.count_if_not_ready):
            m.d.sync += counter.eq(counter + 1)

        with m.If(self.output.ready):
            m.d.sync += self.output.payload.eq(counter)

        return m



from typing import List
from nmigen import *
from nap.util.nmigen_misc import iterator_with_if_elif, nAny
from . import AxiEndpoint

__all__ = ["AxiInterconnect"]


class AxiInterconnect(Elaboratable):
    def __init__(self, upstream):
        """
        A simple single master to many slaves AXI interconnect.

        :type upstream: AxiEndpoint
        :param upstream: The axi master to which the interconnect is connected.
        """
        assert upstream.is_lite, "AXI interconnect only supports AXI lite atm"
        self._upstream = upstream
        self._downstream_ports: List[AxiEndpoint] = []

    def get_port(self):
        """
        Gets a AXI master port connected to the master via this interconnect.

        :return: A new AxiInterface shaped after the upstream port.
        """
        downstream_master = AxiEndpoint.like(self._upstream, name="axi_interconnect_downstream")
        self._downstream_ports.append(downstream_master)
        return downstream_master

    def elaborate(self, platform):
        m = Module()

        for downstream_port in self._downstream_ports:
            m.d.comb += downstream_port.read_address.connect_upstream(self._upstream.read_address)
            m.d.comb += downstream_port.write_address.connect_upstream(self._upstream.write_address)
            m.d.comb += downstream_port.write_data.connect_upstream(self._upstream.write_data)

        # wait until at least one peripherals is ready when writing the addresses
        m.d.comb += self._upstream.read_address.ready.eq(nAny(d.read_address.ready for d in self._downstream_ports))
        m.d.comb += self._upstream.write_address.ready.eq(nAny(d.write_address.ready for d in self._downstream_ports))

        # only one peripheral has to accept written data
        m.d.comb += self._upstream.write_data.ready.eq(nAny(d.write_data.ready for d in self._downstream_ports))

        # we are creating priority encoders here: When multiple peripherals want to answer, we take the answer of the
        # first added peripheral
        for conditional, downstream_port in iterator_with_if_elif(self._downstream_ports, m):
            with conditional(downstream_port.read_data.valid):
                m.d.comb += self._upstream.read_data.connect_upstream(downstream_port.read_data)
        for conditional, downstream_port in iterator_with_if_elif(self._downstream_ports, m):
            with conditional(downstream_port.write_response.valid):
                m.d.comb += self._upstream.write_response.connect_upstream(downstream_port.write_response)

        return m



import numpy as np
import sys
from PIL import Image
from nap.util.plot_util import plt_discrete_hist, plt_image, plt_show

ty = np.int32


def wavelet1d(image, direction_x=False):
    img = image.T if direction_x else image
    cache_dict = {}

    def px(i):
        i2 = i // 2
        if i2 not in cache_dict:
            cache_dict[i2] = np.roll(img, -i2 * 2, 0) if i != 0 else img
        return cache_dict[i2][i % 2::2]

    lf_part = px(0) + px(1)
    hf_part = (px(0) - px(1)) + (-px(-2) - px(-1) + px(2) + px(3) + 4) // 8
    out = np.empty_like(image)
    h, w = out.shape
    if direction_x:
        out[:, :w // 2] = lf_part.T
        out[:, w // 2:] = hf_part.T
    else:
        out[:h // 2] = lf_part
        out[h // 2:] = hf_part
    return out


def inverse_wavelet_1d(image, pad_width=0, direction_x=False):
    img = image.T if direction_x else image
    h, w = img.shape
    lf_part = np.pad(img[:h // 2], pad_width, "edge")
    hf_part = np.pad(img[h // 2:], pad_width, constant_values=0)

    x, y = lf_part.shape
    res = np.zeros((x * 2, y), dtype=ty)
    res[0::2] = (((np.roll(lf_part, +1, 0) - np.roll(lf_part, -1, 0) + 4) >> 3) + hf_part + lf_part) >> 1
    res[1::2] = (((-np.roll(lf_part, +1, 0) + np.roll(lf_part, -1, 0) + 4) >> 3) - hf_part + lf_part) >> 1

    pad_crop = res[2 * pad_width:-2 * pad_width, pad_width:-pad_width] if pad_width > 0 else res
    return pad_crop.T if direction_x else pad_crop


def quantize(image, values, level):
    h, w = image.shape
    parts = [image[:h // 2, :w // 2], image[:h // 2, w // 2:], image[h // 2:, :w // 2], image[h // 2:, w // 2:]]
    for i, (part, value) in enumerate(zip(parts, values)):
        part[:] = np.round(part / value)


def dequantize(image, values, level):
    h, w = image.shape
    parts = [image[:h // 2, :w // 2], image[:h // 2, w // 2:], image[h // 2:, :w // 2], image[h // 2:, w // 2:]]
    for i, (part, value) in enumerate(zip(parts, values)):
        part[:] = part * value


def wavelet2d(image):
    x_transformed = wavelet1d(image, direction_x=True)
    xy_transformed = wavelet1d(x_transformed)
    return xy_transformed


def inverse_wavelet_2d(image, pad_width=0):
    y_transformed = inverse_wavelet_1d(image, pad_width)
    return inverse_wavelet_1d(y_transformed, pad_width, direction_x=True)


def multi_stage_wavelet2d(image, stages, return_all_stages=False, quantization=None):
    h, w = image.shape
    stages_outputs = [image.astype(ty)]
    for i in range(stages):
        transformed = np.copy(stages_outputs[-1])
        transformed[:h // 2 ** i, :w // 2 ** i] = wavelet2d(transformed[:h // 2 ** i, :w // 2 ** i])
        if quantization is not None:
            quantize(transformed[:h // 2 ** i, :w // 2 ** i], quantization[i], i)
        stages_outputs.append(transformed)
    return stages_outputs if return_all_stages else stages_outputs[-1]


def inverse_multi_stage_wavelet2d(image, stages, return_all_stages=False, quantization=None):
    h, w = image.shape
    stages_outputs = [image]
    for i in reversed(range(stages)):
        transformed = np.copy(stages_outputs[-1])
        if quantization is not None:
            dequantize(transformed[:h // 2 ** i, :w // 2 ** i], quantization[i], i)
        transformed[:h // 2 ** i, :w // 2 ** i] = inverse_wavelet_2d(transformed[:h // 2 ** i, :w // 2 ** i])
        stages_outputs.append(transformed)
    return stages_outputs if return_all_stages else stages_outputs[-1]


def compute_psnr(a, b, bit_depth=8):
    diff = a - b
    old_err_state = np.seterr(divide='ignore')
    psnr = 10 * np.log10(((2 ** bit_depth) - 1) ** 2 / np.sum(diff ** 2) * diff.size)
    np.seterr(**old_err_state)
    return psnr


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(f'usage:\n{sys.argv[0]} <input_file>')
        exit(1)
    _, input_file = sys.argv

    if input_file == 'test':
        w = h = 32
        cw = ch = h // 32
        template = np.zeros((2 * ch, 2 * cw), dtype=ty)
        template[:ch, :cw] = 254
        template[ch:, cw:] = 254
        image = np.tile(template, (h // ch, w // ch))
    else:
        image = np.array(Image.open(sys.argv[1])).astype(ty)
    image = image

    quantization = [
        [1, 64, 64, 64],
        [1, 64, 64, 64],
        [1, 64, 64, 64],
    ]

    stages_encode = multi_stage_wavelet2d(image, 3, return_all_stages=True, quantization=quantization)
    stages_decode = inverse_multi_stage_wavelet2d(stages_encode[-1], 3, return_all_stages=True, quantization=quantization)

    plot = True
    for i, (a, b) in enumerate(zip(stages_encode, reversed(stages_decode))):
        crop = 16
        a_crop = a[crop:-crop, crop:-crop]
        b_crop = b[crop:-crop, crop:-crop]
        diff = a_crop - b_crop

        print(f"psnr level {i}: {compute_psnr(a_crop, b_crop)}")

        if plot:
            plt_image(f"lf diff {i}", diff, cmap="bwr", vmin=-2, vmax=+2)
            plt_image(f"lf enc {i}", np.log(a_crop))
            plt_image(f"lf dec {i}", b_crop, cmap="bwr", vmin=-5, vmax=5)

            if i == 0:
                diff_values = a_crop[np.where(diff != 0)]
                plt_discrete_hist(f"diff hist {i}", diff)
                plt_discrete_hist(f"lf hist enc {i}", a_crop)
                plt_discrete_hist(f"lf hist dec {i}", b_crop)
    plt_show()



from nmigen import *

from nap import ControlSignal, StatusSignal
from . import ImageStream

__all__ = ["BlinkDemoVideoSource"]


class BlinkDemoVideoSource(Elaboratable):
    def __init__(self, payload_shape, width, height):
        self.payload_shape = payload_shape
        self.enable = ControlSignal(reset=1)
        self.width = ControlSignal(16, reset=width)
        self.height = ControlSignal(16, reset=height)
        self.frame_counter = StatusSignal(32)
        self.speed = ControlSignal(16, reset=1)

        self.output = ImageStream(payload_shape, name="test_video_source_output")

    def elaborate(self, platform):
        m = Module()

        x_ctr = Signal(16)
        y_ctr = Signal(16)
        m.d.comb += self.output.valid.eq(1)
        m.d.comb += self.output.payload.eq(Repl((self.frame_counter % self.speed) > (self.speed // 2), len(self.output.payload)))
        with m.If(self.output.ready):
            with m.If(x_ctr < self.width):
                m.d.sync += x_ctr.eq(x_ctr + 1)
            with m.Else():
                m.d.sync += x_ctr.eq(0)
                m.d.comb += self.output.line_last.eq(1)
                with m.If(y_ctr < self.height):
                    m.d.sync += y_ctr.eq(y_ctr + 1)
                with m.Else():
                    m.d.sync += y_ctr.eq(0)
                    m.d.comb += self.output.frame_last.eq(1)
                    with m.If(self.enable):
                        m.d.sync += self.frame_counter.eq(self.frame_counter + 1)
                    with m.Else():
                        m.d.sync += x_ctr.eq(x_ctr)
                        m.d.comb += self.output.valid.eq(0)

        return m



import marshal, sys, types
from itertools import repeat
from pathlib import Path
import numpy as np
import rawpy
from multiprocessing import Pool

from .dng import read_dng, write_dng
from .py_compressor import NumericRange, to_chunks, rle_compress_chunks, compute_symbol_frequencies, generate_huffman_tables, merge_symbol_frequencies, get_huffman_size
from .py_wavelet import inverse_multi_stage_wavelet2d, multi_stage_wavelet2d, ty
from .vifp import vifp_mscale

levels = 3
bit_depth = None


def load_image(path):
    images = {}
    filename = Path(f).stem

    red, green1, green2, blue, order, estimated_bit_depth = read_dng(f)

    images[f'{filename}--R'] = red
    images[f'{filename}--G1'] = green1
    images[f'{filename}--G2'] = green2
    images[f'{filename}--B'] = blue

    # images with different bit depths cant be combined in one run
    global bit_depth
    assert bit_depth is None or bit_depth == estimated_bit_depth
    bit_depth = estimated_bit_depth

    return images, {(f'{filename}--R', f'{filename}--G1', f'{filename}--G2', f'{filename}--B'): order}


def wrapper(args):
    fun, g, *args = args
    code = marshal.loads(fun)
    fun = types.FunctionType(code, dict(g) | globals(), "fn")
    return fun(*args)


def pmap(fn, *iterables, capture=(), threads=None):
    g = [(k, v) for k, v in globals().items() if k in capture]
    with Pool(threads) as pool:
        return pool.map(wrapper, zip(repeat(marshal.dumps(fn.__code__)), repeat(g), *iterables))


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f'usage:\n{sys.argv[0]} input...')
        exit(1)
    _, *files = sys.argv

    images = {}
    metadata = {}
    for f in files:
        new_images, new_metadata = load_image(f)
        images.update(new_images)
        metadata.update(new_metadata)

    input_range = NumericRange(0, 2 ** bit_depth - 1)

    quantization = np.array([
        [1, 48, 48, 72],
        [2, 48, 48, 24],
        [1, 48, 48, 24],
    ], dtype=ty)


    def each_transform_rle(img):
        filename, image = img
        transformed = multi_stage_wavelet2d(image, levels, quantization=quantization)

        chunks = list(to_chunks(transformed, levels))
        region_codes, uncompressed_chunks = zip(*chunks)
        rle_chunks = list(rle_compress_chunks(chunks, levels, input_range, quantization))

        symbol_frequencies = compute_symbol_frequencies(region_codes, rle_chunks, levels, input_range, quantization)

        roundtripped = inverse_multi_stage_wavelet2d(transformed, levels, quantization=quantization)

        return filename, region_codes, rle_chunks, symbol_frequencies, image, roundtripped


    filenames, region_codes_array, rle_chunks_array, symbol_frequencies_array, original, roundtripped = \
        zip(*map(each_transform_rle, images.items()))

    huffman_tables = generate_huffman_tables(merge_symbol_frequencies(symbol_frequencies_array), levels, input_range, quantization)


    def recombine_images(metadata):
        rggb_filenames, order = metadata
        filename = rggb_filenames[0].split("--")[0]

        per_plane = []
        for f in rggb_filenames:
            idx = filenames.index(f)
            per_plane.append((original[idx], roundtripped[idx], region_codes_array[idx], rle_chunks_array[idx]))
        return per_plane, order, filename


    recombined_images = map(recombine_images, metadata.items())


    def each_compute_vifp_ratio(recombined_image):
        (r, g1, g2, b), order, filename = recombined_image
        original_path = write_dng(f'{filename}-original', r[0], g1[0], g2[0], b[0], bit_depth, order)
        debayered_original = rawpy.imread(original_path).postprocess(output_bps=16)
        roundtripped_path = write_dng(f'{filename}-roundtripped', r[1], g1[1], g2[1], b[1], bit_depth, order)
        debayered_roundtripped = rawpy.imread(roundtripped_path).postprocess(output_bps=16)
        vifp = vifp_mscale(debayered_original, debayered_roundtripped)

        compressed_sizes = []
        for plane in (r, g1, g2, b):
            original, roundtripped, region_codes, rle_chunks = plane
            huffman_encoded_size = get_huffman_size(huffman_tables, region_codes, rle_chunks, levels, input_range, quantization)
            compressed_sizes.append((original.size * bit_depth) / huffman_encoded_size)

        print(f'{filename: <30}\t1:{np.mean(compressed_sizes):02f}\tvif: {vifp:02f}')


    list(pmap(each_compute_vifp_ratio, recombined_images, capture=('bit_depth', 'order', 'quantization', 'huffman_tables'), threads=4))



import unittest
from nmigen import *

from nap.stream.formal_util import verify_stream_output_contract, LegalStreamSource
from nap.stream.sim_util import write_to_stream, read_from_stream
from nap.cores.compression.bit_stuffing import BitStuffer, VariableWidthStream
from nap.util.sim import SimPlatform


class BitStufferTest(unittest.TestCase):
    def test_basic(self):
        platform = SimPlatform()
        m = Module()

        input_data = ['00000001', '00000011', '00000111', '00001111']

        input = VariableWidthStream(32)
        bit_stuffer = m.submodules.bit_stuffer = BitStuffer(input, 32)

        def write_process():
            for word in input_data:
                yield from write_to_stream(input, payload=int(word, 2), current_width=len(word))

        def read_process():
            read = "{:032b}".format((yield from read_from_stream(bit_stuffer.output)))
            print(read)
            self.assertEquals(read, "".join(reversed(input_data)))

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_advanced(self):
        platform = SimPlatform()
        m = Module()

        input_data = ['1110010', '1101001', '1111101', '1111101', '000110', '11010111000101', '0100111110010011', '011100', '000110', '001110', '1111101', '1100101', '0100111110010011',
                      '111111101101', '111111101111111']

        input = VariableWidthStream(32)
        bit_stuffer = m.submodules.bit_stuffer = BitStuffer(input, 32)

        def write_process():
            for word in input_data:
                yield from write_to_stream(input, payload=int(word, 2), current_width=len(word))

        def read_process():
            read = []
            while True:
                try:
                    read.append("{:032b}".format((yield from read_from_stream(bit_stuffer.output))))
                except TimeoutError:
                    break
            read_bitstring = "".join("".join(reversed(x)) for x in read)
            print(read_bitstring)
            input_bitstring = "".join("".join(reversed(x)) for x in input_data)
            print(input_bitstring)
            print(read_bitstring)
            self.assertTrue(input_bitstring.startswith(read_bitstring))

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_advanced_last(self):
        platform = SimPlatform()
        m = Module()

        input_data = ['1110010', '1101001', '1111101', '1111101', '000110', '11010111000101', '0100111110010011', '011100', '000110', '001110', '1111101', '1100101', '0100111110010011',
                      '111111101101', '111111101111111']

        input = VariableWidthStream(32)
        bit_stuffer = m.submodules.bit_stuffer = BitStuffer(input, 32)

        def write_process():
            for i, word in enumerate(input_data):
                yield from write_to_stream(input, payload=int(word, 2), current_width=len(word), last=(i == len(input_data) - 1))

        def read_process():
            read = []
            for i in range(100):
                v, last = (yield from read_from_stream(bit_stuffer.output, extract=("payload", "last")))
                read.append("{:032b}".format(v))
                if last:
                    break
            assert i != 99, "timeout"
            read_bitstring = "".join("".join(reversed(x)) for x in read)
            print(read_bitstring)
            input_bitstring = "".join("".join(reversed(x)) for x in input_data)
            print(input_bitstring)
            print(read_bitstring)
            self.assertTrue(read_bitstring.startswith(input_bitstring))

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_output_stream_properties(self):
        input = VariableWidthStream(32)
        verify_stream_output_contract(BitStuffer(input, 32), support_modules=(LegalStreamSource(input),))



import unittest
from nmigen.sim import Passive
from nap import SimPlatform, BasicStream, write_to_stream, read_from_stream
from nap.stream.formal_util import verify_stream_output_contract, LegalStreamSource
from nap.cores.axi import AxiEndpoint, answer_read_burst
from .stream_reader import AxiReader, AxiReaderBurster


class TestAxiReader(unittest.TestCase):
    def test_basic(self):
        platform = SimPlatform()

        memory = {i: i + 100 for i in range(1000)}
        read_sequence = [
            0,
            100,
            108,
            116,
            2,
            7,
            *[i + 100 for i in range(0, 400, 8)],
            0,
            100,
            108,
            116,
            2,
            7,
        ]
        golden_read_result = [memory[addr] for addr in read_sequence]

        axi = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)
        address_stream = BasicStream(32)
        dut = AxiReader(address_stream, axi)

        def write_address_process():
            for addr in read_sequence:
                yield from write_to_stream(address_stream, payload=addr)
        platform.add_process(write_address_process, "sync")

        def read_data_process():
            read_result = []
            while len(read_result) < len(golden_read_result):
                read = yield from read_from_stream(dut.output)
                read_result.append(read)
            self.assertEqual(read_result, golden_read_result)
        platform.add_process(read_data_process, "sync")

        def axi_answer_process():
            yield Passive()
            while True:
                yield from answer_read_burst(axi, memory)
        platform.add_process(axi_answer_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_reader_stream_output(self):
        axi = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)
        verify_stream_output_contract(
            AxiReader(BasicStream(32), axi),
            support_modules=(LegalStreamSource(axi.read_data),)
        )

    def test_burster_stream_output(self):
        i = BasicStream(32)
        verify_stream_output_contract(
            AxiReaderBurster(i),
            support_modules=(LegalStreamSource(i),)
        )



import random
import unittest
from nmigen import *
from nmigen.sim import Passive
from nap import SimPlatform, do_nothing, BasicStream, PacketizedStream, read_from_stream, write_to_stream
from nap.stream import LegalStreamSource, verify_stream_output_contract
from nap.cores import BufferedSyncStreamFIFO, StreamTee, StreamGearbox, SimpleStreamGearbox


class TestGearbox(unittest.TestCase):
    def test_gearbox_7_to_3(self):
        input = BasicStream(7)
        dut = StreamGearbox(input, 3)

        def writer():
            yield from write_to_stream(input, payload=0b0_010_001)
            yield from write_to_stream(input, payload=0b10_011_10)
            yield from write_to_stream(input, payload=0b000_111_1)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output)), 0b001)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b010)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b100)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b011)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b110)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b111)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b000)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_3_to_7(self):
        input = BasicStream(3)
        dut = StreamGearbox(input, 7)

        def writer():
            yield from write_to_stream(input, payload=0b001)
            yield from write_to_stream(input, payload=0b010)
            yield from write_to_stream(input, payload=0b100)
            yield from write_to_stream(input, payload=0b011)
            yield from write_to_stream(input, payload=0b110)
            yield from write_to_stream(input, payload=0b111)
            yield from write_to_stream(input, payload=0b000)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output)), 0b0_010_001)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b10_011_10)
            self.assertEqual((yield from read_from_stream(dut.output)), 0b000_111_1)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_4_to_8_last(self):
        input = PacketizedStream(4)
        dut = StreamGearbox(input, 8)

        def writer():
            yield from write_to_stream(input, payload=0b0001, last=0)
            yield from write_to_stream(input, payload=0b0010, last=1)
            yield from write_to_stream(input, payload=0b0100, last=0)
            yield from write_to_stream(input, payload=0b1000, last=0)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0010_0001, 1))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b1000_0100, 0))

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_8_to_4_last(self):
        input = PacketizedStream(8)
        dut = StreamGearbox(input, 4)

        def writer():
            yield from write_to_stream(input, payload=0b00_10_00_01, last=1)
            yield from write_to_stream(input, payload=0b10_00_01_00, last=0)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0001, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0010, 1))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0100, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b1000, 0))

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_48_to_12_last(self):
        input = PacketizedStream(48)
        dut = StreamGearbox(input, 12)

        def writer():
            yield from write_to_stream(input, payload=0b000000000100_000000000010_000000000001_100000000000, last=1)
            yield from write_to_stream(input, payload=0b000001000000_000000100000_000000010000_000000001000, last=0)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b100000000000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000001, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000010, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000100, 1))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000001000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000010000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000100000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000001000000, 0))

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_12_to_48_to_64(self):
        m = Module()
        platform = SimPlatform()
        input = BasicStream(12)

        tee = m.submodules.tee = StreamTee(input)
        gear_12_to_48 = m.submodules.gear_12_to_48 = StreamGearbox(tee.get_output(), 48)
        gear_48_to_64 = m.submodules.gear_48_to_64 = StreamGearbox(gear_12_to_48.output, 64)

        gear_12_to_64 = m.submodules.gear_12_to_64 = StreamGearbox(tee.get_output(), 64)

        def writer():
            yield Passive()
            random.seed(0)
            while True:
                yield from write_to_stream(input, payload=random.randrange(0, 2**12))
        platform.add_process(writer, "sync")

        def reader():
            for _ in range(100):
                a = yield from read_from_stream(gear_12_to_64.output)
                b = yield from read_from_stream(gear_48_to_64.output)
                print(f"{a:064b}")
                self.assertEqual(a, b)
        platform.add_process(reader, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(m)

    def test_dont_loose_data(self):
        input = BasicStream(16)
        dut = StreamGearbox(input, 8)

        def writer():
            for i in range(0, 100, 2):
                yield from write_to_stream(input, payload=(((i + 1) << 8) | i))
                if i % 7 == 0:
                    yield from do_nothing()

        def reader():
            for i in range(100):
                got = (yield from read_from_stream(dut.output, extract="payload"))
                print(got)
                self.assertEqual(got, i)
                if i % 3 == 0:
                    yield from do_nothing()

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_dont_loose_last_8_to_4(self):
        input = PacketizedStream(8)
        dut = StreamGearbox(input, 4)

        def writer():
            last_count_gold = 0
            for i in range(50):
                last = (i % 5 == 0)
                last_count_gold += last
                yield from write_to_stream(input, payload=0, last=last)
                if i % 7 == 0:
                    yield from do_nothing()
            self.assertEqual(last_count_gold, 10)

        def reader():
            last_count = 0
            for i in range(100):
                last = (yield from read_from_stream(dut.output, extract="last"))
                last_count += last
                if i % 10 == 1:
                    assert last
                else:
                    assert not last
                if i % 3 == 0:
                    yield from do_nothing()
            self.assertEquals(last_count, 10)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_dont_loose_last_16_to_4(self):
        input = PacketizedStream(16)
        dut = StreamGearbox(input, 4)

        def writer():
            last_count_gold = 0
            for i in range(50):
                last = (i % 5 == 0)
                last_count_gold += last
                yield from write_to_stream(input, payload=0, last=(i % 5 == 0))
                if i % 3 == 0:
                    yield from do_nothing()
            self.assertEqual(last_count_gold, 10)

        def reader():
            last_count = 0
            for i in range(200):
                last_count += (yield from read_from_stream(dut.output, extract="last"))
                if i % 10 == 0:
                    yield from do_nothing()
            self.assertEquals(last_count, 10)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_gearbox_automated_fifo(self):
        def string_generator(string, chunk_size):
            last = 0
            while True:
                to_yield = ""
                for _ in range(chunk_size):
                    try:
                        to_yield += string[last]
                    except IndexError:
                        return
                    last += 1
                yield "".join(reversed(to_yield))

        def gold_gen(input_width, output_width, output_amount=10):
            bit_sequence = ""
            for i in range(output_amount):
                bit_sequence += "{{:0{}b}}".format(input_width).format(i % (2 ** input_width))
            return (
                [int(s, 2) for s in string_generator(bit_sequence, input_width)],
                [int(s, 2) for s in string_generator(bit_sequence, output_width)]
            )

        def test_gearbox(input_width, output_width):
            input = PacketizedStream(input_width)
            m = Module()
            fifo_in = m.submodules.fifo_in = BufferedSyncStreamFIFO(input, 100)
            gearbox = m.submodules.gearbox = StreamGearbox(fifo_in.output, output_width)
            fifo_out = m.submodules.fifo_out = BufferedSyncStreamFIFO(gearbox.output, 100)

            input_data, output_data = gold_gen(input_width, output_width)

            def writer():
                for v in input_data:
                    yield from write_to_stream(input, payload=v)

            def reader():
                for i, v in enumerate(output_data):
                    read = (yield from read_from_stream(fifo_out.output))
                    self.assertEqual(read, v)

            platform = SimPlatform()
            platform.add_sim_clock("sync", 100e6)
            platform.add_process(writer, "sync")
            platform.add_process(reader, "sync")
            platform.sim(m)

        test_gearbox(8, 4)
        test_gearbox(7, 3)
        test_gearbox(48, 12)

        for input_width in range(1, 100, 7):
            for output_width in range(1, 100, 3):
                test_gearbox(input_width, output_width)

    def test_output_stream_contract(self):
        input = BasicStream(7)
        verify_stream_output_contract(StreamGearbox(input, 3))


class TestSimpleGearbox(unittest.TestCase):
    def test_simple_gearbox_8_to_4_last(self):
        input = PacketizedStream(8)
        dut = SimpleStreamGearbox(input, 4)

        def writer():
            yield from write_to_stream(input, payload=0b00_10_00_01, last=1)
            yield from write_to_stream(input, payload=0b10_00_01_00, last=0)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0001, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0010, 1))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b0100, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b1000, 0))

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_simple_gearbox_48_to_12_last(self):
        input = PacketizedStream(48)
        dut = SimpleStreamGearbox(input, 12)

        def writer():
            yield from write_to_stream(input, payload=0b000000000100_000000000010_000000000001_100000000000, last=1)
            yield from write_to_stream(input, payload=0b000001000000_000000100000_000000010000_000000001000, last=0)

        def reader():
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b100000000000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000001, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000010, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000000100, 1))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000001000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000010000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000000100000, 0))
            self.assertEqual((yield from read_from_stream(dut.output, extract=("payload", "last"))), (0b000001000000, 0))

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_simple_gearbox_dont_loose_last_16_to_4(self):
        input = PacketizedStream(16)
        dut = SimpleStreamGearbox(input, 4)

        def writer():
            last_count_gold = 0
            for i in range(50):
                last = (i % 5 == 0)
                last_count_gold += last
                yield from write_to_stream(input, payload=0, last=(i % 5 == 0))
                if i % 3 == 0:
                    yield from do_nothing()
            self.assertEqual(last_count_gold, 10)

        def reader():
            last_count = 0
            for i in range(200):
                last_count += (yield from read_from_stream(dut.output, extract="last"))
                if i % 10 == 0:
                    yield from do_nothing()
            self.assertEquals(last_count, 10)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(writer, "sync")
        platform.add_process(reader, "sync")
        platform.sim(dut)

    def test_output_stream_contract(self):
        input = BasicStream(8)
        verify_stream_output_contract(SimpleStreamGearbox(input, 4), support_modules=(LegalStreamSource(input),))



# TODO: this is the beginning of a full hdmi implementation. Finish this.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum


class BitVec:
    def __init__(self, len):
        pass


class PacketConstructable(ABC):
    @abstractmethod
    def lower(self):
        raise NotImplementedError()


@dataclass
class DataIslandPacket(PacketConstructable, ABC):
    header_bytes: list
    data_bytes: list


class RgbOrYcbcr(Enum):
    RGB = 0b00
    YCBCR422 = 0b01
    YCBCR444 = 0b10


class ActiveFormatInformationPresent(Enum):
    NO_DATA = 0
    INFORMATION_VALID = 1


class BarInfo(Enum):
    NOT_VALID = 0b00
    VERTICAL_BAR_INFO_VALID = 0b01
    HORIZONTAL_BAR_INFO_VALID = 0b10
    BOTH_VALID = 0b11


class ScanInformation(Enum):
    NO_DATA = 0b00
    OVERSCANNED = 0b01
    UNDERSCANNED = 0b10


class Colorimetry(Enum):
    NO_DADA = 0b00
    SMPTE170_OR_ITU601 = 0b01
    ITU709 = 0b10
    EXTENDED = 0b11


class PictureAspectRatio(Enum):
    NO_DATA = 0b00
    ASPECT_4_3 = 0b01
    ASPECT_16_9 = 0b10


class ActiveFormatAspectRatio(Enum):
    SAME_AS_PICTURE_ASPECT = 0b1000


class ItContent(Enum):
    NO_DATA = 0
    IT_CONTENT = 1


class ExtendedColorimetry(Enum):
    XVYCC601 = 0b000
    XVYCC709 = 0b001


class RgbQuantizationRange(Enum):
    DEFAULT = 0b00
    LIMITED = 0b01
    FULL = 0b10


class NonUniformPictureScaling(Enum):
    NO_KNOWN = 0b00
    HORIZONTAL = 0b01
    VERTICAL = 0b10
    BOTH = 0b11


class VideoIdentificationCode(Enum):
    NO_INFORMATION = 0
    VID_1920_1080_60p = 16
    VID_1920_1080_30p = 34
    VID_1920_1080_25p = 33
    VID_1920_1080_24p = 32
    VID_1920_1080_50p = 31

@dataclass
class AviInfoFrame(PacketConstructable):
    # header
    packet_type = 0x82
    info_frame_version = 0x02
    info_frame_lengh = 0x0D

    # bady
    scan_information: ScanInformation
    rgb_or_ycbcr: RgbOrYcbcr
    active_format_information_present: ActiveFormatInformationPresent
    bar_info: BarInfo
    colorimetry: Colorimetry
    picture_aspect_ratio: PictureAspectRatio


    def lower(self):
        return DataIslandPacket(
            header_bytes=[self.packet_type, self.info_frame_version, self.info_frame_lengh]
        )



# An experiment that allows to flash the USB3 plugin module via JTAG on the micro via bitbanging and
# MMIO GPIO.
from nmigen import *
from nap import *
from nap.vendor.xilinx_s7 import Pll


class Top(Elaboratable):
    def elaborate(self, platform: ZynqSocPlatform):
        usb3_plugin_connect(platform, "south")

        m = Module()

        platform.ps7.fck_domain(20e6, "fclk_in")
        pll = m.submodules.pll = Pll(20e6, 40, 1, input_domain="fclk_in")
        pll.output_domain("bitclk", 2)
        pll.output_domain("sync", 8)

        clocking = m.submodules.clocking = ClockingDebug("fclk_in", "bitclk", "sync")

        usb3_plugin = platform.request("usb3_plugin", "south")

        if isinstance(platform, MicroR2Platform):
            m.submodules.mmio_gpio = MmioGpio([
                usb3_plugin.jtag.tms,
                usb3_plugin.jtag.tck,
                usb3_plugin.jtag.tdi,
                usb3_plugin.jtag.tdo,

                usb3_plugin.jtag_enb,
                usb3_plugin.program,
                usb3_plugin.init,
                usb3_plugin.done,
            ])

        counter = m.submodules.counter = CounterStreamSource(32)
        m.submodules.tx = PluginModuleStreamerTx(usb3_plugin.lvds, counter.output, bitclk_domain="bitclk")

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform,), possible_socs=(ZynqSocPlatform,))



from nmigen import *
from nmigen_boards.zturn_lite_z010 import ZTurnLiteZ010Platform


class IlaPlatform:
    def __init__(self, platform):
        self.platform = platform

        self.probes = {}

    # we pass through all platform methods, because we pretend to be one
    def __getattr__(self, item):
        return getattr(self.platform, item)

    def probe(self, signal, name):
        self.probes[name] = signal
        return signal


if __name__ == "__main__":
    class Top(Elaboratable):
        def __init__(self):
            pass

        def elaborate(self, platform):
            m = Module()

            counter = platform.probe(Signal(32), name="counter")
            m.d.sync += counter.eq(counter + 1)

            # add some ila core here

            return m


    platform = IlaPlatform(ZTurnLiteZ010Platform)
    platform.build(Top())


import unittest
from nap import SimPlatform
from . import SocPlatform


class ConcreteSocPlatform(SocPlatform):
    pass

class TestSocPlatform(unittest.TestCase):
    def test_is_instance(self):
        platform = ConcreteSocPlatform(SimPlatform())
        self.assertIsInstance(platform, SimPlatform)
        self.assertIsInstance(platform, ConcreteSocPlatform)
        self.assertIsInstance(platform, SocPlatform)



from nmigen import *
from nap import ImageStream, DOWNWARDS, BufferedSyncStreamFIFO
from nap.cores.video import ImageSplitter, ImageCombiner, BlackLineGenerator, ImageConvoluter


class Wavelet1D(Elaboratable):
    def __init__(self, input: ImageStream, width, height, direction_y):
        self.input = input
        self.output = ImageStream(input.payload.shape(), name="wavelet_1D_output")

        self.height = height
        self.width = width
        self.direction_y = direction_y

    def elaborate(self, platform):
        m = Module()

        def transformer_function(x, y, image_proxy):
            output = Signal.like(image_proxy[x, y])

            def px(shift):
                if self.direction_y:
                    return image_proxy[x, y + shift]
                else:
                    return image_proxy[x + shift, y]

            # even pixels are lf while odd pixels are hf
            with m.If((y if self.direction_y else x) % 2 == 0):
                m.d.comb += output.eq((px(0) + px(1)) // 2)
            with m.Else():
                m.d.comb += output.eq(((px(0) - px(1) + (-px(-2) - px(-1) + px(2) + px(3)) // 8) // 2) + (2**len(self.input.payload) // 2))
            return output

        video_transformer = m.submodules.video_transformer = ImageConvoluter(self.input, transformer_function, self.width, self.height)
        m.d.comb += self.output.connect_upstream(video_transformer.output)

        return m


class Wavelet2D(Elaboratable):
    def __init__(self, input: ImageStream, width, height):
        self.input = input
        self.output = ImageStream(input.payload.shape(), name="wavelet_2D_output")

        self.height = height
        self.width = width

    def elaborate(self, platform):
        m = Module()

        wavelet_x = m.submodules.wavelet_x = Wavelet1D(self.input, self.width, self.height, direction_y=False)
        wavelet_y = m.submodules.wavelet_y = Wavelet1D(wavelet_x.output, self.width, self.height, direction_y=True)
        m.d.comb += self.output.connect_upstream(wavelet_y.output)

        return m


class MultiStageWavelet2D(Elaboratable):
    """Does a multi level wavelet transform while producing a wide image that has the lf part on the left and the hf parts on the right"""
    def __init__(self, input: ImageStream, width, height, stages, level=1):
        self.input = input
        self.output = input.clone(name="wavelet_level{}_output".format(level))
        self.output.is_hf = Signal() @ DOWNWARDS

        self.width = width
        self.level = level
        self.height = height
        self.stages = stages
        self.fifos = []

    def elaborate(self, platform):
        m = Module()

        def calculate_needed_bottom_buffer(stage):
            if stage == 1:
                return 0
            else:
                return 6 * (self.width // 2**(stage - 1)) + calculate_needed_bottom_buffer(stage - 1)


        def x_preroll(width, stages):
            if stages == 1:
                return 0
            elif stages == 2:
                return width
            else:
                return width * 3 // 4 + x_preroll(width // 2, stages - 1)

        transformer = m.submodules.transformer = Wavelet2D(self.input, self.width, self.height)
        splitter = m.submodules.splitter = ImageSplitter(transformer.output, self.width, self.height)
        for i, output in enumerate(splitter.outputs):
            output.is_hf = Signal(reset=(i != 0)) @ DOWNWARDS

        lf_output = splitter.outputs[0]
        lf_processed = splitter.outputs[0].clone()
        if self.stages == 1:
            m.d.comb += lf_processed.connect_upstream(lf_output)
            m.d.comb += lf_processed.is_hf.eq(0)
        else:
            next_stage_input = lf_output.clone()
            m.d.comb += next_stage_input.connect_upstream(lf_output)
            self.next_stage = next_stage = m.submodules.next_stage = MultiStageWavelet2D(next_stage_input, self.width // 2, self.height // 2, self.stages - 1, self.level + 1)
            padding_generator = m.submodules.padding_generator = BlackLineGenerator(lf_output.payload.shape(), x_preroll(self.width, self.stages), black_value=0) # (2**len(self.input.payload) // 2))

            n_preroll_lines = 5 # TODO(robin): why is it exactly this value
            preroll_lines = Signal(range(n_preroll_lines + 1))
            even_odd = Signal(reset = 1)
            output_counting_stream = lf_processed
            new_line = output_counting_stream.line_last & output_counting_stream.valid & output_counting_stream.ready
            with m.If(preroll_lines < n_preroll_lines):
                with m.If(new_line):
                    m.d.sync += preroll_lines.eq(preroll_lines + 1)
                m.d.comb += lf_processed.connect_upstream(padding_generator.output, allow_partial=True)
            with m.Else():
                with m.If(new_line):
                    m.d.sync += even_odd.eq(~even_odd)
                with m.If(even_odd):
                    m.d.comb += lf_processed.connect_upstream(next_stage.output)
                with m.Else():
                    m.d.comb += lf_processed.connect_upstream(padding_generator.output, allow_partial=True)

        hf_top_right_fifo = m.submodules.hf_topright_fifo = BufferedSyncStreamFIFO(splitter.outputs[1], self.width // 2 - 1)
        hf_combiner = m.submodules.hf_combiner = ImageCombiner(hf_top_right_fifo.output, *splitter.outputs[2:], interleave=True, output_name="hf_combiner_output")
        output_combiner = m.submodules.output_combiner = ImageCombiner(lf_processed, hf_combiner.output, interleave=False, output_name="output_combiner_output")

        if self.level == 1:
            output_buffer_size = 0
        else:
            output_buffer_size = { 2: 5 * self.width // 2 - 4, 1: self.width * 2 - 4 }[self.stages]
        output_fifo = m.submodules.output_fifo = BufferedSyncStreamFIFO(output_combiner.output, output_buffer_size)
        m.d.comb += self.output.connect_upstream(output_fifo.output)

        self.fifos.append(output_fifo)

        return m



import unittest
from nap import AxiEndpoint, axil_read, axil_write, CsrBank, ControlSignal, ZynqSocPlatform, SimPlatform, do_nothing


class TestAxiSlave(unittest.TestCase):
    def check_csr_bank(self, num_csr=10, testdata=0x12345678, use_axi_interconnect=False):
        platform = ZynqSocPlatform(SimPlatform(), use_axi_interconnect)
        csr_bank = CsrBank()
        for i in range(num_csr):
            csr_bank.reg("csr#{}".format(i), ControlSignal(32))

        def testbench():
            axi = platform.axi_lite_master
            for addr in [0x4000_0000 + (i * 4) for i in range(num_csr)]:
                yield from axil_read(axi, addr)
                yield from axil_write(axi, addr, testdata)
                self.assertEqual(testdata, (yield from axil_read(axi, addr)))

        platform.sim(csr_bank, (testbench, "axi_lite"))

    def test_csr_bank_aggregator(self):
        self.check_csr_bank(use_axi_interconnect=False)

    def test_csr_bank_interconnect(self):
        self.check_csr_bank(use_axi_interconnect=True)

    def test_simple_test_csr_bank(self):
        platform = ZynqSocPlatform(SimPlatform())
        csr_bank = CsrBank()
        csr_bank.reg("csr", ControlSignal(32))

        def testbench():
            axi: AxiEndpoint = platform.axi_lite_master
            yield axi.read_address.payload.eq(0x4000_0000)
            yield axi.read_address.valid.eq(1)
            yield from do_nothing()

        platform.sim(csr_bank, (testbench, "axi_lite"))
